# coding=utf-8
import collections
import csv
import numpy as np
import tensorflow as tf
from py_common_util.common.annotations import print_exec_time
from tensorflow_datasets.core import api_utils
import tensorflow_datasets as tfds
from tensorflow.python import feature_column
from py_common_util.common.common_utils import CommonUtils
from py_common_util.tensorflow.tf_utils import TFUtils

_CITATION = """\
@ONLINE {super_dataset,
author = "Frank E. Harrell Jr., Thomas Cason",
title  = "super_dataset dataset",
month  = "oct",
year   = "2017",
url    = "https://www.openml.org/d/40945"
}
"""

_DESCRIPTION = ("Dataset describing the survival status of "
                "individual passengers on the super_dataset. Missing values in "
                "the original dataset are represented using ?. "
                "Float and int missing values are replaced with -1, string "
                "missing values are replaced with 'Unknown'.")

_EMBARKED_DICT = collections.OrderedDict([
    ("C", "Cherbourg"), ("Q", "Queenstown"), ("S", "Southampton"),
    ("?", "Unknown")
])

_PCLASS_DICT = collections.OrderedDict([
    ("1", "1st_class"), ("2", "2nd_class"), ("3", "3rd_class")
])

_SURVIVED_DICT = {"1": "survived", "0": "died"}


class DefaultDatasetBuilderConfig(tfds.core.BuilderConfig):
  """BuilderConfig for SuperDatasetBuilder."""

  @api_utils.disallow_positional_args
  def __init__(self, text_encoder_config=None, **kwargs):
    """BuilderConfig for SuperDatasetBuilder.
    Args:
      text_encoder_config: `tfds.features.text.TextEncoderConfig`, configuration
        for the `tfds.features.text.TextEncoder` used for the SuperDatasetBuilder `"text"`
        feature.
      **kwargs: keyword arguments forwarded to super.
    """
    super().__init__(**kwargs)
    self.text_encoder_config = (
        text_encoder_config or tfds.features.text.TextEncoderConfig())


class SuperDatasetBuilder(tfds.core.GeneratorBasedBuilder):
    """
    tensorflow dataset builder https://www.tensorflow.org/datasets/catalog/overview
    """
    BUILDER_CONFIGS = [
        DefaultDatasetBuilderConfig(
            name="plain_text",
            version="0.0.1",
            description="Plain text",
        ),
        DefaultDatasetBuilderConfig(
            name="bytes",
            version="0.0.1",
            description=("Uses byte-level text encoding with "
                         "`tfds.features.text.ByteTextEncoder`"),
            text_encoder_config=tfds.features.text.TextEncoderConfig(
                encoder=tfds.features.text.ByteTextEncoder()),
        ),
        DefaultDatasetBuilderConfig(
            name="subwords8k",
            version="0.0.1",
            description=("Uses `tfds.features.text.SubwordTextEncoder` with 8k "
                         "vocab size"),
            text_encoder_config=tfds.features.text.TextEncoderConfig(
                encoder_cls=tfds.features.text.SubwordTextEncoder,
                vocab_size=2 ** 13),
        ),
        DefaultDatasetBuilderConfig(
            name="subwords32k",
            version="0.0.1",
            description=("Uses `tfds.features.text.SubwordTextEncoder` with "
                         "32k vocab size"),
            text_encoder_config=tfds.features.text.TextEncoderConfig(
                encoder_cls=tfds.features.text.SubwordTextEncoder,
                vocab_size=2 ** 15),
        ),
    ]

    def __init__(self, data_dir=None, config=None):
        super().__init__(data_dir=data_dir, config=config)
        self._num_shards = 3
        self._train_data_file = "/Users/tony/tensorflow_datasets/manual_downloads/phpMYEkMl.csv"
        self._test_data_file = None
        self._validation_data_file = None

    def _info(self):
        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict({
                "label": tfds.features.ClassLabel(names=["died", "survived"]),
                "feature": {name: dtype
                            for name, (dtype, func) in DEMO_FEATURE_DICT.items()},
            }),
            supervised_keys=("feature", "label"),
            urls=["https://xxx"],
            citation=_CITATION
        )

    def _split_generators(self, dl_manager):
        generator_array = []
        if self._train_data_file:
            generator_array.append(tfds.core.SplitGenerator(
                name=tfds.Split.TRAIN,
                num_shards=3,
                gen_kwargs={
                    "data_file": self._train_data_file
                }))
        if self._test_data_file:
            generator_array.append(tfds.core.SplitGenerator(
                name=tfds.Split.TEST,
                num_shards=1,
                gen_kwargs={
                    "data_file": self._test_data_file
                }))
        if self._validation_data_file:
            generator_array.append(tfds.core.SplitGenerator(
                name=tfds.Split.VALIDATION,
                num_shards=1,
                gen_kwargs={
                    "data_file": self._validation_data_file
                }))
        return generator_array

    def _generate_examples(self, **kwargs):
        """Generate features and target given the directory path.
        Args:
          file_path: path where the csv file is stored
        Yields:
          The features and the target
        """
        with tf.io.gfile.GFile(self._train_data_file) as f:
            raw_data = csv.DictReader(f)
            for row in raw_data:
                survive_val = row.pop("survived")
                yield {
                    "label": SuperDatasetBuilder.convert_to_label(survive_val, _SURVIVED_DICT),
                    "feature": {
                        name: DEMO_FEATURE_DICT[name][1](value)
                        for name, value in row.items()
                    }
                }

    @staticmethod
    def build_dataset_builder_config(name="plain_text", version="0.0.1", description="", text_encoder_config=None):
        return DefaultDatasetBuilderConfig(name=name, version=version, description=description, text_encoder_config=text_encoder_config)

    @staticmethod
    def convert_to_float(d):
      return -1.0 if d == "?" else np.float32(d)

    @staticmethod
    def convert_to_float64(d):
      return -1.0 if d == "?" else np.float64(d)

    @staticmethod
    def convert_to_int(d):
      return -1 if d == "?" else np.int32(d)  # np.int64(d) is serializable

    @staticmethod
    def convert_to_int64(d):
      return -1 if d == "?" else np.int64(d)  # np.int64(d) is serializable

    @staticmethod
    def convert_to_string(d):
      return "Unknown" if d == "?" else d

    @staticmethod
    def convert_to_label(d, dictionary):
      return dictionary[d]

    @staticmethod
    def return_same(d):
      return d

    @staticmethod
    def list_builders():
        return tfds.list_builders()

    @staticmethod
    def convert_to_int_from_string(string, dtype=np.int64, count=-1, sep=','):
        """
        @:param sep ','或'|'
        #>>> np.fromstring('8,9', dtype=np.int64,  sep=',')
        array([8, 9])
        """
        if len(string.split("|")) > 1:
            sep = "|"
        return np.fromstring(string, dtype=dtype, count=count, sep=sep)

    @staticmethod
    def convert_to_float32_from_string(string, dtype=np.float32, count=-1, sep=','):
        return SuperDatasetBuilder.convert_to_int_from_string(string, dtype, count, sep)

    @staticmethod
    def convert_to_float64_from_string(string, dtype=np.float64, count=-1, sep=','):
        return SuperDatasetBuilder.convert_to_int_from_string(string, dtype, count, sep)

    @staticmethod
    @print_exec_time
    def build_and_download_and_prepare(dataset_builder_cls_name=None, config_name="plain_text"):
        dataset_builder = tfds.builder(dataset_builder_cls_name + "/" + config_name)
        dataset_builder.download_and_prepare()
        return dataset_builder.info

    @staticmethod
    @print_exec_time
    def load_batched_datasets(dataset_builder_cls_name=None,
                             config_name=None,
                             split=[tfds.Split.ALL],
                             map_fn_feature_label=None,
                             batch_size:int=128,
                             shuffle:int=1000,
                             prefetch:int=10,
                             epochs:int=1) -> list:
        """
        注意：多次重新生成不同版本的数据的时候，可能需要先要手工文件夹：~/tensorflow_datasets/<<dataset_builder_cls_name>>/
        :param dataset_builder_cls_name: e.g. "mnist"
        :param config_name: e.g. "plain_text"
        :param split:
        :param map_fn_feature_label:
        e.g. def map_fn_feature_label(feature, label):
                feature = feature/255
                feature = tf.squeeze(feature)  # tf.squeeze删除大小是1的维度
                return feature, label
        :param shuffle: <1不shuffle
        :param batch_size: -1为全量数据, 或者如：SuperDatasetBuilder.get_num_examples("fashion_mnist", tfds.Split.TEST)
        :param prefetch:
        :param epochs:
        :return: list[tf.data.Dataset]:
        e.g. shape: [(num_batch, batch_size, feature_size, label_size)],
             data: [num_batch个[(feature, label),...batch_size个...,(feature, label)]]
        usage:
        batched_datasets = SuperDatasetBuilder.load_batched_datasets("recommend_dataset_builder", config_name="bytes", split=[tfds.Split.TRAIN], batch_size=128)
        for dataset in batched_datasets:
            for data in SuperDatasetBuilder.read_batched_features(dataset):
                print(data[0]["pclass"], data[1])
        # or
        for batched_examples in tfds.as_numpy(batched_datasets):
            batched_features, batched_labels = batched_examples # 对应info.supervised_keys的元素个数
        """
        dataset_name = (dataset_builder_cls_name + "/" + config_name) if config_name else dataset_builder_cls_name
        if not batch_size or batch_size < 1:
            batch_size = 1
        datasets, data_info = tfds.load(dataset_name,
                                    split=split,
                                    as_supervised=True,
                                    with_info=True,
                                    as_dataset_kwargs=dict(
                                        shuffle_files=False,
                                    ))
        batched_datasets = []  # 对应split的数据集个数
        for dataset in datasets:
            if map_fn_feature_label:
                dataset = dataset.map(map_fn_feature_label)
            if shuffle < 1:
                batched_datasets.append(dataset.batch(batch_size).prefetch(prefetch).repeat(epochs))
            else:
                batched_datasets.append(dataset.shuffle(shuffle).batch(batch_size).prefetch(prefetch).repeat(epochs))
        # if len(batched_datasets) == 1:
        #     # 1个元素的列表就直接返回该元素
        #     batched_datasets = batched_datasets[0]
        # return batched_datasets
        concated_batched_datasets = None
        for data in batched_datasets:
            if concated_batched_datasets is None:
                concated_batched_datasets = data
            else:
                concated_batched_datasets = concated_batched_datasets.concatenate(data)
        return concated_batched_datasets

    @staticmethod
    def read_batched_datasets(batched_datasets):
        """
        batched_datasets,  shape (num_batch, (batch_size个(feature, label)), e.g. (100, (128, ((28,28), (1))))
        ```
        for batched_data in MovieLensDatasetBuilder.read_batched_datasets(test_dataset):
            features, labels = batched_data
        ```
        :param batched_datasets:
        :return:
        """
        for datasets in tfds.as_numpy(batched_datasets):
            features, labels = datasets
            yield features, labels

    @staticmethod
    def split_full_datasets(full_datasets: tf.data.Dataset, num_examples=0, train_size=0.7, validation_size=0, test_size=0.3):
        """
        按比例切割数据集
        或依赖sklearn工具: from sklearn.model_selection import train_test_split
        # 采用 80 - 20 的比例切分训练集和验证集
        # input_tensor_train, input_tensor_val, target_tensor_train, target_tensor_val = train_test_split(input_tensor, target_tensor, test_size=0.2)
        :param full_datasets 开发调试时可以取部分数据，e.g. 让full_datasets=full_datasets.take(10)
        :param num_examples, e.g. total num examples | total num batched
        :param train_size e.g. 0.7 或 0.1
        :param validation_size e.g. 0.1 或 0
        :param test_size e.g. 0.2 或 0
        ```
            train_data, validation_data, test_data = MovieLensDatasetBuilder.split_full_datasets(full_datasets,
                                                                                                  num_examples=MovieLensDatasetBuilder.get_total_num_batch(
                                                                                                      num_examples,
                                                                                                      128),
                                                                                                  train_size=0.7,
                                                                                                  validation_size=0.1,
                                                                                                  test_size=0.2)
        ```
        """
        train_shard_num = int(train_size*10)
        train_size = int(train_size * num_examples)
        validation_shard_num = int(validation_size*10)
        validation_size = int(validation_size * num_examples)
        test_shard_num = int(test_size*10)
        test_size = int(test_size * num_examples)

        # 注意：用skip/take方法取validationh和test数据集很慢
        # train_dataset = full_datasets.take(train_size)
        # validation_dataset = full_datasets.skip(train_size).take(validation_size)
        # test_dataset = full_datasets.skip(train_size).skip(validation_size).take(test_size)

        # 注意：用skip/take方法取validationh和test数据集很慢
        # train_dataset = full_datasets.take(train_size)
        # test_dataset = full_datasets.skip(train_size)
        # validation_dataset = test_dataset.skip(validation_size)
        # test_dataset = test_dataset.take(test_size)
        # 用shard方法获取数据集比较快，注意要确保shard之后再去做shuffle的操作

        train_dataset = full_datasets.take(0)
        validation_dataset = full_datasets.take(0)
        test_dataset = full_datasets.take(0)
        for i in range(train_shard_num):
            train_dataset = train_dataset.concatenate(full_datasets.shard(10, i))
        for i in range(train_shard_num, train_shard_num + validation_shard_num):
            validation_dataset = validation_dataset.concatenate(full_datasets.shard(10, i))
        for i in range(train_shard_num + validation_shard_num, train_shard_num + validation_shard_num + test_shard_num):
            test_dataset = test_dataset.concatenate(full_datasets.shard(10, i))
        return train_dataset, validation_dataset, test_dataset

    @staticmethod
    def camelcase_to_snakecase(class_name):
        """
        Convert camel-case string to snake-case.
        usage:
        ```
        camelcase_to_snakecase(MovieLensDatasetBuilder.__name__)
        ```
        """
        return CommonUtils.camelcase_to_snakecase(class_name)

    @staticmethod
    def get_num_examples(dataset_builder_cls_name=None, split=tfds.Split.TRAIN):
        """
        获取指定数据集的长度
        :param dataset_builder_cls_name: e.g. "mnist/plain_text" 或 "mnist"
        :param split:
        :return:
        """
        return tfds.builder(dataset_builder_cls_name).info.splits[split].num_examples

    @staticmethod
    def get_total_num_batch(num_total, batch_size, allow_smaller_final_batch=True):
        return TFUtils.total_num_batch(num_total, batch_size, allow_smaller_final_batch)

    @staticmethod
    def get_info(dataset_builder_cls_name=None):
        """dataset_builder_cls_name: mnist/plain_text 或 mnist"""
        return tfds.builder(dataset_builder_cls_name).info

    @staticmethod
    def get_split_name_list(dataset_builder_cls_name=None):
        return list(tfds.builder(dataset_builder_cls_name).info.splits.keys())

    @staticmethod
    def build_all_numeric_feature_columns(dataset_builder_cls_name=None, supervised_keys_feature_name="feature"):
        """
        for tf.estimator https://www.tensorflow.org/guide/feature_columns?hl=zh-cn
        查看feature_columns的值：tf.keras.layers.DenseFeatures(feature_columns)(example).numpy()
        or tf.feature_column.input_layer(features, [feature_column]).numpy()
        构建分布式Tensorflow模型系列:特征工程 https://zhuanlan.zhihu.com/p/41663141
        :param dataset_builder_cls_name: e.g. "mnist"
        :param supervised_keys_feature_name: e.g. "image"
        :return:
        """
        feature_columns = []
        feature_types = SuperDatasetBuilder.get_info(dataset_builder_cls_name).features.dtype[supervised_keys_feature_name]
        if isinstance(feature_types, tf.dtypes.DType):
            if feature_types.is_integer or feature_types.is_floating:
                feature_columns.append(feature_column.numeric_column(supervised_keys_feature_name, dtype=feature_types))
        else:
            for feature_name in feature_types:
                feature_type = feature_types[feature_name]
                if feature_type.is_integer or feature_type.is_floating:
                    feature_columns.append(SuperDatasetBuilder.build_numeric_column(feature_name, dtype=feature_type))
        return feature_columns

    @staticmethod
    def build_numeric_feature_columns(feature_name_list, dtype=tf.float32):
        """
        for tf.estimator https://www.tensorflow.org/guide/feature_columns?hl=zh-cn
        查看feature_columns的值：tf.keras.layers.DenseFeatures(feature_columns)(example).numpy()
        or tf.feature_column.input_layer(features, [feature_column]).numpy()
        构建分布式Tensorflow模型系列:特征工程 https://zhuanlan.zhihu.com/p/41663141
        :param feature_name_list: e.g. ["user_id", "item_id]
        :param dtype:
        :return:
        """
        return [feature_column.numeric_column(feature_name, dtype=dtype) for feature_name in feature_name_list]

    @staticmethod
    def build_linear_model(features_value, combined_feature_columns=[]):
        """
        for tf.estimator https://www.tensorflow.org/guide/feature_columns?hl=zh-cn
        查看feature_columns的值：tf.keras.layers.DenseFeatures(feature_columns)(example).numpy()
        or tf.feature_column.input_layer(features, [feature_column]).numpy()
        构建分布式Tensorflow模型系列:特征工程 https://zhuanlan.zhihu.com/p/41663141
        线性模型LinearModel
        对所有特征进行线性加权操作（数值和权重值相乘）。
        ```
        from tensorflow.python import feature_column
        featrues = {
            'price': [[1.0], [5.0], [10.0]],
            'color': [['R'], ['G'], ['B']]
        }
        price_column = feature_column.numeric_column('price')
        color_column = feature_column.categorical_column_with_vocabulary_list('color', ['R', 'G', 'B'])
        prediction = SuperDatasetBuilder.build_linear_model(featrues, [price_column, color_column]).numpy()
        print("feature_values" + "***"*40 + "\n", prediction)
        结果为：
        feature_values************************************************************************************************************************
        [[0.]
        [0.]
        [0.]]
        ```
        :param features_value: dict or tf.dataset
        :param combined_feature_columns:
        :return:
        """
        return feature_column.linear_model(features_value, combined_feature_columns)

    @staticmethod
    def build_weighted_categorical_column(categorical_column, weight_feature_name):
        """
        for tf.estimator https://www.tensorflow.org/guide/feature_columns?hl=zh-cn
        查看feature_columns的值：tf.keras.layers.DenseFeatures(feature_columns)(example).numpy()
        or tf.feature_column.input_layer(features, [feature_column]).numpy()
        构建分布式Tensorflow模型系列:特征工程 https://zhuanlan.zhihu.com/p/41663141
        权重分类列WeightedCategoricalColumn
        权重改变了one-hot模式，不仅包含0或1，还带有权重值, e.g.
        [array([[1., 0., 0., 0.],
        [0., 0., 0., 5.],
        [0., 4., 0., 0.],
        [0., 0., 8., 0.],
        [3., 0., 0., 0.]], dtype=float32)]
        ```
        from tensorflow.python import feature_column
        features = {'color': [['R'], ['A'], ['G'], ['B'], ['R']],
                    'weight': [[1.0], [5.0], [4.0], [8.0], [3.0]]}
        color_f_c = feature_column.categorical_column_with_vocabulary_list(
            'color', ['R', 'G', 'B', 'A'], dtype=tf.string, default_value=-1
        )
        column = feature_column.weighted_categorical_column(color_f_c, 'weight')
        indicator = feature_column.indicator_column(column)
        feature_values = feature_column.input_layer(features, [indicator]).numpy()
        print("feature_values" + "***"*40 + "\n", feature_values)
        结果如下：
        feature_values************************************************************************************************************************
        [[1. 0. 0. 0.]
        [0. 0. 0. 5.]
        [0. 4. 0. 0.]
        [0. 0. 8. 0.]
        [3. 0. 0. 0.]]
        ```
        :param categorical_column:
        :param weight_feature_name:
        :return:
        """
        return feature_column.weighted_categorical_column(categorical_column, weight_feature_name)

    @staticmethod
    def build_one_hot_feature_column(categorical_column):
        """
        支持one_hot和multi-hot两种模式：i.e. If my input is ["bob", "wanda"] or [0, 2], I want to get a [1, 0, 1] vector and not [[1, 0, 0], [0, 0, 1]]
        for tf.estimator https://www.tensorflow.org/guide/feature_columns?hl=zh-cn
        查看feature_columns的值：tf.keras.layers.DenseFeatures(feature_columns)(example).numpy()
        or tf.feature_column.input_layer(features, [feature_column]).numpy()
        构建分布式Tensorflow模型系列:特征工程 https://zhuanlan.zhihu.com/p/41663141
        指标列
        指标列和嵌入列从不直接处理特征，而是将分类列视为输入。
        使用指标列时，我们指示 TensorFlow 完成我们在分类 product_class 样本中看到的确切操作。
        也就是说，指标列将每个类别视为独热矢量中的一个元素，其中匹配类别的值为 1，其余类别为 0：0->[1,0,0,0]
        ```python
        name = indicator_column(categorical_column_with_vocabulary_list(
          'name', ['bob', 'george', 'wanda'])
        columns = [name, ...]
        features = tf.parse_example(..., features=make_parse_example_spec(columns))
        dense_tensor = input_layer(features, columns)
        dense_tensor == [[1, 0, 0]]  # If "name" bytes_list is ["bob"]
        dense_tensor == [[1, 0, 1]]  # If "name" bytes_list is ["bob", "wanda"]
        dense_tensor == [[2, 0, 0]]  # If "name" bytes_list is ["bob", "bob"]
        ```
        :param categorical_column:
        :return:
        """
        return feature_column.indicator_column(categorical_column)

    @staticmethod
    def build_categorical_column_with_identity(feature_name, num_buckets, default_value=None):
        """
        for tf.estimator https://www.tensorflow.org/guide/feature_columns?hl=zh-cn
        查看feature_columns的值：tf.keras.layers.DenseFeatures(feature_columns)(example).numpy()
        or tf.feature_column.input_layer(features, [feature_column]).numpy()
        构建分布式Tensorflow模型系列:特征工程 https://zhuanlan.zhihu.com/p/41663141
        ```
        分类标识列
        可将分类标识列视为分桶列的一种特殊情况。在传统的分桶列中，每个分桶表示一系列值（例如，从 1960 年到 1979 年）。
        在分类标识列中，每个分桶表示一个唯一整数。例如，假设您想要表示整数范围 [0, 4)。也就是说，您想要表示整数 0、1、2 或 3。在这种情况下，分类标识映射如下所示：
        0->[1,0,0,0]
        1->[0,1,0,0]
        2->[0,0,1,0]
        3->[0,0,0,1]
        与分桶列一样，模型可以在分类标识列中学习每个类别各自的权重。例如，我们使用唯一的整数值来表示每个类别，而不是使用某个字符串来表示 product_class。即：
        0="kitchenware"
        1="electronics"
        2="sport"
        调用 tf.feature_column.categorical_column_with_identity 以实现类别标识列。例如：
        # Create categorical output for an integer feature named "my_feature_b",
        # The values of my_feature_b must be >= 0 and < num_buckets
        identity_feature_column = tf.feature_column.categorical_column_with_identity(
            key='my_feature_b',
            num_buckets=4) # Values [0, 4)
        # In order for the preceding call to work, the input_fn() must return
        # a dictionary containing 'my_feature_b' as a key. Furthermore, the values
        # assigned to 'my_feature_b' must belong to the set [0, 4).
        def input_fn():
            ...
            return ({ 'my_feature_a':[7, 9, 5, 2], 'my_feature_b':[3, 1, 2, 2] },
                    [Label_values])
        测试代码
        import tensorflow as tf
        test = {'cat': [1,3,2,0,3]} #index分别为1,3,2,0,3
        column = tf.feature_column.categorical_column_with_identity(
            key='cat',
            num_buckets=4)
        indicator = tf.feature_column.indicator_column(column)
        tensor = tf.feature_column.input_layer(test, [indicator])
        print(tensor.numpy())
        # 输出结果
        array([[0., 1., 0., 0.],
               [0., 0., 0., 1.],
               [0., 0., 1., 0.],
               [1., 0., 0., 0.],
               [0., 0., 0., 1.]], dtype=float32)
        ```
        :param feature_name:
        :param num_buckets:
        :param default_value:
        :return:
        """
        return feature_column.categorical_column_with_identity(key=feature_name, num_buckets=num_buckets, default_value=default_value)

    @staticmethod
    def build_embedding_feature_column(categorical_column, embedding_dimension):
        """
        ref: 看Google如何实现Wide & Deep模型（2.1） https://zhuanlan.zhihu.com/p/47965313
            https://www.tensorflow.org/api_docs/python/tf/feature_column/embedding_column
        for tf.estimator https://www.tensorflow.org/guide/feature_columns?hl=zh-cn
        看Google如何实现Wide & Deep模型（2.2）https://zhuanlan.zhihu.com/p/47970601
        查看feature_columns的值：tf.keras.layers.DenseFeatures(feature_columns)(example).numpy()
        构建分布式Tensorflow模型系列:特征工程 https://zhuanlan.zhihu.com/p/41663141
        or tf.feature_column.input_layer(features, [feature_column]).numpy()
        嵌入列
        现在，假设我们有一百万个可能的类别，或者可能有十亿个，而不是只有三个。出于多种原因，随着类别数量的增加，使用指标列来训练神经网络变得不可行。
        我们可以使用嵌入列来克服这一限制。嵌入列并非将数据表示为很多维度的one-hot矢量，而是将数据表示为低维度普通矢量，其中每个单元格可以包含任意数字，
        而不仅仅是 0 或 1。通过使每个单元格能够包含更丰富的数字，嵌入列包含的单元格数量远远少于指标列。
        我们来看一个将指标列和嵌入列进行比较的示例。假设我们的输入样本包含多个不同的字词（取自仅有 81 个字词的有限词汇表）。
        我们进一步假设数据集在 4 个不同的样本中提供了下列输入字词：
        "dog"
        "spoon"
        "scissors"
        "guitar"
        处理样本时，其中一个 categorical_column_with... 函数会将样本字符串映射到分类数值。例如，一个函数将“spoon”映射到 [32]。
        （32 是我们想象出来的，实际值取决于映射函数。）然后，您可以通过下列两种方式之一表示这些分类数值：
        作为指标列。函数将每个分类数值转换为一个 81 元素的矢量（因为我们的词汇表由 81 个字词组成），
        将 1 置于分类值 (0, 32, 79, 80) 的索引处，将 0 置于所有其他位置。
        作为嵌入列。函数将分类数值 (0, 32, 79, 80) 用作对照表的索引。该对照表中的每个槽位都包含一个 3 元素矢量。
        e.g. indexs (0, 32, 79, 80) -> embedding: [[0.421,0.399,0.512],[0.312,0.632,0.412],[0.294,0.112,0.578],[0.722,0.689,0.219]]
        ```
        color_data = {"color": ["R", "B"]}
        color_column = SuperDatasetBuilder.build_categorical_column_with_vocabulary_list("color", ["R", "G", "B"])
        color_embeding = SuperDatasetBuilder.build_embedding_feature_column(color_column, 8)
        color_embeding_value = feature_column.input_layer(color_data, [color_embeding]).numpy()
        print("embedding" + "*" * 40)
        print(color_embeding_value)
        结果为：
        embedding****************************************
        [[ 0.12240461  0.41843808  0.6384718  -0.16248055  0.00317956 -0.10421363
           0.01167523 -0.11598229]
         [ 0.19527617  0.3898876   0.3911946  -0.6085922  -0.3114762   0.55712754
           0.13116577 -0.6390011 ]]
        ```
        :param categorical_column: 可以来自build_categorical_column_with_identity()
        :param embedding_dimension: 假定number_of_categories为81, 则embedding_dimension一般为：math.ceil(number_of_categories**0.25) 请注意，这只是一个一般规则；您可以根据需要设置嵌入维度的数量。
        :return:
        """
        return feature_column.embedding_column(categorical_column=categorical_column, dimension=embedding_dimension)

    @staticmethod
    def build_categorical_column_with_vocabulary_list(feature_name, vocabulary_list, default_value=0):
        """
        for tf.estimator https://www.tensorflow.org/guide/feature_columns?hl=zh-cn
        查看feature_columns的值：tf.keras.layers.DenseFeatures(feature_columns)(example).numpy()
        or tf.feature_column.input_layer(features, [feature_column]).numpy()
        构建分布式Tensorflow模型系列:特征工程 https://zhuanlan.zhihu.com/p/41663141
        categorical_column_with_vocabulary_list 根据明确的词汇表将每个字符串映射到一个整数。
        :param feature_name: e.g. age
        :param vocabulary_list: 有顺序的类别列表 e.g. ('male', 'female') or [1,2,3,4,5]
        :return:
        """
        return feature_column.categorical_column_with_vocabulary_list(feature_name, vocabulary_list, default_value=default_value)

    @staticmethod
    def build_categorical_column_with_vocabulary_file(feature_name, vocabulary_file, vocabulary_size=None, default_value=0):
        """
        for tf.estimator https://www.tensorflow.org/guide/feature_columns?hl=zh-cn
        查看feature_columns的值：tf.keras.layers.DenseFeatures(feature_columns)(example).numpy()
        or tf.feature_column.input_layer(features, [feature_column]).numpy()
        构建分布式Tensorflow模型系列:特征工程 https://zhuanlan.zhihu.com/p/41663141
        当词汇表很长时，需要输入的内容太多了。对于此类情况，
        请改为调用 tf.feature_column.categorical_column_with_vocabulary_file，
        以便将词汇放在单独的文件中。例如：
        # Given input "feature_name_from_input_fn" which is a string,
        # create a categorical feature to our model by mapping the input to one of
        # the elements in the vocabulary file
        vocabulary_feature_column =
            tf.feature_column.categorical_column_with_vocabulary_file(
                key=feature_name_from_input_fn,
                vocabulary_file="product_class.txt",
                vocabulary_size=3)
        product_class.txt 中的每个词汇元素应各占一行。在我们的示例中：
        kitchenware
        electronics
        sports
        ```
        feature_value_dict = {'letter': [["A's", 'B']]}
        categorical_feature_column = SuperDatasetBuilder.build_categorical_column_with_vocabulary_list(
            "letter", ["A's", "A", "B", "C"])
        one_hot_column = SuperDatasetBuilder.build_one_hot_column(categorical_feature_column)
        feature_values = tf.keras.layers.DenseFeatures(one_hot_column)(feature_value_dict).numpy()
        print(feature_values)
        ```
        :param feature_name: e.g. sex
        :param vocabulary_file
        :param vocabulary_size 为None是即为vocabulary_file文件的长度
        :param default_value
        :return:
        """
        return feature_column.categorical_column_with_vocabulary_file(feature_name, vocabulary_file, vocabulary_size=vocabulary_size, default_value=default_value)

    @staticmethod
    def build_categorical_column_with_hash_bucket(feature_name, hash_buckets_size=1):
        """
        for tf.estimator https://www.tensorflow.org/guide/feature_columns?hl=zh-cn
        查看feature_columns的值：tf.keras.layers.DenseFeatures(feature_columns)(example).numpy()
        or tf.feature_column.input_layer(features, [feature_column]).numpy()
        构建分布式Tensorflow模型系列:特征工程 https://zhuanlan.zhihu.com/p/41663141
        经过哈希处理的列
        到目前为止，我们处理的示例都包含很少的类别。例如，我们的 product_class 示例只有 3 个类别。
        但是通常，类别的数量非常大，以至于无法为每个词汇或整数设置单独的类别，因为这会消耗太多内存。
        对于此类情况，我们可以反问自己：“我愿意为我的输入设置多少类别？”实际上，tf.feature_column.categorical_column_with_hash_bucket 函数使您能够指定类别的数量。
        对于这种类型的特征列，模型会计算输入的哈希值，然后使用模运算符将其置于其中一个 hash_bucket_size 类别中，如以下伪代码所示：
        # pseudocode
        feature_id = hash(raw_feature) % hash_buckets_size
        创建 feature_column 的代码可能如下所示：
        hashed_feature_column =
            tf.feature_column.categorical_column_with_hash_bucket(
                key = "some_feature",
                hash_buckets_size = 100) # The number of categories
        此时，您可能会认为：“这太疯狂了！”，这种想法很正常。毕竟，我们是将不同的输入值强制划分成更少数量的类别。
        这意味着，两个可能不相关的输入会被映射到同一个类别，这样一来，神经网络也会面临同样的结果。下图显示了这一困境，
        即厨具和运动用品都被分配到类别（哈希分桶）12：
        与机器学习中的很多反直觉现象一样，事实证明哈希技术经常非常有用。这是因为哈希类别为模型提供了一些分隔方式。
        模型可以使用其他特征进一步将厨具与运动用品分隔开来。
        :param feature_name:
        :param hash_buckets_size: The number of categories e.g. 100
        :return:
        """
        hashed_feature_column = feature_column.categorical_column_with_hash_bucket(feature_name, hash_buckets_size=hash_buckets_size)
        return hashed_feature_column

    @staticmethod
    def build_bucketized_feature_column(numeric_feature_column, boundaries):
        """
        for tf.estimator https://www.tensorflow.org/guide/feature_columns?hl=zh-cn
        查看feature_columns的值：tf.keras.layers.DenseFeatures(feature_columns)(example).numpy()
        or tf.feature_column.input_layer(features, [feature_column]).numpy()
        构建分布式Tensorflow模型系列:特征工程 https://zhuanlan.zhihu.com/p/41663141
        分桶列：e.g. bucketize the numeric column on the years 1960, 1980, and 2000.
        如，将年份数据分成四个分桶，模型将按以下方式表示这些分桶：
        日期范围	表示为…
        < 1960 年	[1, 0, 0, 0]
        >= 1960 年但 < 1980 年	[0, 1, 0, 0]
        >= 1980 年但 < 2000 年	[0, 0, 1, 0]
        >= 2000 年	[0, 0, 0, 1]
        :param numeric_feature_column: e.g. tf.feature_column.numeric_column("Year")
        :param boundaries: e.g. [1960, 1980, 2000]
        :return:
        """
        bucketized_feature_column = feature_column.bucketized_column(
            source_column=numeric_feature_column,
            boundaries=boundaries)
        return bucketized_feature_column

    @staticmethod
    def build_crossed_feature_column(feature_columns=None, hash_bucket_size=10):
        """
        for tf.estimator https://www.tensorflow.org/guide/feature_columns?hl=zh-cn
        查看feature_columns的值：tf.keras.layers.DenseFeatures(feature_columns)(example).numpy()
        or tf.feature_column.input_layer(features, [feature_column]).numpy()
        构建分布式Tensorflow模型系列:特征工程 https://zhuanlan.zhihu.com/p/41663141
        组合列
        通过将多个特征组合为一个特征（称为特征组合），模型可学习每个特征组合的单独权重。
        更具体地说，假设我们希望模型计算佐治亚州亚特兰大的房产价格。这个城市的房产价格在不同位置差异很大。
        在确定对房产位置的依赖性方面，将纬度和经度表示为单独的特征用处不大；但是，将纬度和经度组合为一个特征则可精确定位位置。
        假设我们将亚特兰大表示为一个 100x100 的矩形网格区块，按纬度和经度的特征组合标识全部 10000 个区块。借助这种特征组合，
        模型可以针对与各个区块相关的房价条件进行训练，这比单独的经纬度信号强得多。
        crossed_column 仅构建 hash_bucket_size 参数所请求的数字，而不是构建这个可能非常庞大的输入表。
        特征列通过在输入元组上运行哈希函数，然后使用 hash_bucket_size 进行模运算，为索引分配一个样本。
        如前所述，执行哈希函数和模函数会限制类别的数量，但会导致类别冲突；也就是说，多个（纬度、经度）特征组合最终位于同一个哈希分桶中。
        但实际上，执行特征组合对于模型的学习能力仍具备重大价值。
        有些反直觉的是，在创建特征组合时，通常仍应在模型中包含原始（未组合）特征（如前面的代码段中所示）。
        独立的纬度和经度特征有助于模型区分组合特征中发生哈希冲突的样本。
        For example, if the input features are:
        * SparseTensor referred by first key:
        ```python
        shape = [2, 2]
        {
            [0, 0]: "a"
            [1, 0]: "b"
            [1, 1]: "c"
        }
        ```
        * SparseTensor referred by second key:
        ```python
        shape = [2, 1]
        {
            [0, 0]: "d"
            [1, 0]: "e"
        }
        ```
        then crossed feature will look like:
        ```python
        shape = [2, 2]
        {
          [0, 0]: Hash64("d", Hash64("a")) % hash_bucket_size
          [1, 0]: Hash64("e", Hash64("b")) % hash_bucket_size
          [1, 1]: Hash64("e", Hash64("c")) % hash_bucket_size
        }
        ```
        :param feature_columns:
        :param hash_bucket_size:
        :return:
        """
        return feature_column.crossed_column(feature_columns, hash_bucket_size)


DEMO_FEATURE_DICT = collections.OrderedDict([
    ("pclass", (tfds.features.ClassLabel(names=_PCLASS_DICT.values()),
                lambda d: SuperDatasetBuilder.convert_to_label(d, _PCLASS_DICT))),
    ("name", (tf.string, SuperDatasetBuilder.convert_to_string)),
    ("sex", (tfds.features.ClassLabel(names=["male", "female"]), SuperDatasetBuilder.return_same)),
    ("age", (tf.float32, SuperDatasetBuilder.convert_to_float)),
    ("sibsp", (tf.int32, SuperDatasetBuilder.convert_to_int)),
    ("parch", (tf.int32, SuperDatasetBuilder.convert_to_int)),
    ("ticket", (tf.string, SuperDatasetBuilder.convert_to_string)),
    ("fare", (tf.float32, SuperDatasetBuilder.convert_to_float)),
    ("cabin", (tf.string, SuperDatasetBuilder.convert_to_string)),
    ("embarked", (tfds.features.ClassLabel(names=_EMBARKED_DICT.values()),
                  lambda d: SuperDatasetBuilder.convert_to_label(d, _EMBARKED_DICT))),
    ("boat", (tf.string, SuperDatasetBuilder.convert_to_string)),
    ("body", (tf.int32, SuperDatasetBuilder.convert_to_int)),
    ("home.dest", (tf.string, SuperDatasetBuilder.convert_to_string))
])

if __name__ == '__main__':
    batched_datasets = SuperDatasetBuilder.load_batched_datasets("ted_hrlr_translate",
                                                                 config_name="ru_to_en",
                                                                 split=[tfds.Split.TRAIN])
    batched_dataset_features = SuperDatasetBuilder.read_batched_datasets(batched_datasets)
    for features, labels in batched_dataset_features:
        print(f"features={''.join([feature.decode('utf-8') for feature in features])}, labels={''.join([label.decode('utf-8') for label in labels])}")
    print(f"end!!")



# coding=utf-8
import sys
import collections
import tensorflow as tf
import tensorflow_datasets as tfds
from py_common_util.common.annotations import Override
from skydl.datasets.super_tf_dataset_builder import SuperDatasetBuilder, DefaultDatasetBuilderConfig
from skydl.models_zoo.bert.official.nlp.bert.tokenization import FullTokenizer, printable_text, convert_to_unicode


class ChineseBertNerDatasetBuilder(SuperDatasetBuilder):
    """
    中文bert模型用的ner数据集
    参考：https://github.com/ProHiryu/bert-chinese-ner
    https://github.com/huggingface/transformers/blob/master/examples/run_tf_ner.py
    https://github.com/kamalkraj/BERT-NER-TF
    https://github.com/ymcui/Chinese-BERT-wwm
    https://github.com/liloi/bert-tf2
    """
    BUILDER_CONFIGS = [
        DefaultDatasetBuilderConfig(
            name="plain_text",
            version="0.0.1",
            description="Plain text",
        )
    ]
    # set value of max_seq_length
    MAX_SEQ_LENGTH = 128
    # feature column names
    INPUT_IDS_COLUMN = "input_ids"    # input_ids
    INPUT_MASK_COLUMN = "input_mask"  # input_mask
    SEGMENTS_ID_COLUMN = "segments_id"  # segments_id
    # target column name
    TARGET_COLUMN = "target"
    # used to change labels to ids
    LABEL_DICT = collections.OrderedDict([
        (i, i) for i in ["O", "B-PER", "I-PER", "B-ORG", "I-ORG", "B-LOC", "I-LOC", "X", "[CLS]", "[SEP]"]
    ])
    # feature column dict
    FEATURE_DICT = collections.OrderedDict([
        ("input_ids", (tfds.features.Tensor(shape=(MAX_SEQ_LENGTH,), dtype=tf.int32), SuperDatasetBuilder.convert_to_int)),
        ("input_mask", (tfds.features.Tensor(shape=(MAX_SEQ_LENGTH,), dtype=tf.int32), SuperDatasetBuilder.convert_to_int)),
        ("segment_ids", (tfds.features.Tensor(shape=(MAX_SEQ_LENGTH,), dtype=tf.int32), SuperDatasetBuilder.convert_to_int)),
        # 在_generate_examples()里单独处理了target列，这里不需要重复定义
    ])

    def __init__(self, data_dir=None, config=None):
        super().__init__(data_dir=data_dir, config=config)
        self._num_shards = 1
        self._train_data_file = sys.path[0] + "/data/bert/ner/train.txt"
        self._test_data_file = sys.path[0] + "/data/bert/ner/test.txt"
        self._validation_data_file = sys.path[0] + "/data/bert/ner/dev.txt"

    @Override(SuperDatasetBuilder)
    def _info(self):
        return tfds.core.DatasetInfo(
            builder=self,
            description=("Chinese Bert-based Named Entity Recognition dataset"),
            features=tfds.features.FeaturesDict({
                "label": tfds.features.Tensor(shape=(self.MAX_SEQ_LENGTH,), dtype=tf.int32),
                "feature": {name: dtype
                            for name, (dtype, func) in self.FEATURE_DICT.items()},
            }),
            supervised_keys=("feature", "label"),
            urls=["https://github.com/huggingface/transformers/issues/1166"],
            citation="@ONLINE {Chinese Bert-based Named Entity Recognition dataset}"
        )

    @Override(SuperDatasetBuilder)
    def _split_generators(self, dl_manager):
        generator_array = []
        if self._train_data_file:
            generator_array.append(tfds.core.SplitGenerator(
                name=tfds.Split.TRAIN,
                num_shards=3,
                gen_kwargs={
                    "max_seq_length": self.MAX_SEQ_LENGTH,
                    "data_type": "train",
                    "data_path": self._train_data_file
                }))
        if self._test_data_file:
            generator_array.append(tfds.core.SplitGenerator(
                name=tfds.Split.TEST,
                num_shards=1,
                gen_kwargs={
                    "max_seq_length": self.MAX_SEQ_LENGTH,
                    "data_type": "test",
                    "data_path": self._test_data_file
                }))
        if self._validation_data_file:
            generator_array.append(tfds.core.SplitGenerator(
                name=tfds.Split.VALIDATION,
                num_shards=1,
                gen_kwargs={
                    "max_seq_length": self.MAX_SEQ_LENGTH,
                    "data_type": "validation",
                    "data_path": self._validation_data_file
                }))
        return generator_array

    @classmethod
    def readfile(cls, filename):
        '''
        read file
        '''
        f = open(filename)
        data = []
        sentence = []
        label = []
        for line in f:
            if len(line) == 0 or line.startswith('-DOCSTART') or line[0] == "\n":
                if len(sentence) > 0:
                    data.append((sentence, label))
                    sentence = []
                    label = []
                continue
            splits = line.split(' ')
            sentence.append(splits[0])
            #label.append(splits[-1][:-1])
            label.append(splits[-1].replace("\n", ""))
        if len(sentence) > 0:
            data.append((sentence, label))
            sentence = []
            label = []
        return data

    @classmethod
    def readfile2(cls, input_file):
        """Reads a BIO data."""
        with open(input_file) as f:
            lines = []
            words = []
            labels = []
            for line in f:
                contends = line.strip()
                word = line.strip().split(' ')[0]
                label = line.strip().split(' ')[-1]
                if contends.startswith("-DOCSTART-"):
                    words.append('')
                    continue
                # if len(contends) == 0 and words[-1] == '。':
                if len(contends) == 0:
                    l = ' '.join([label for label in labels if len(label) > 0])
                    w = ' '.join([word for word in words if len(word) > 0])
                    lines.append([l, w])
                    words = []
                    labels = []
                    continue
                words.append(word)
                labels.append(label)
            return lines

    @Override(SuperDatasetBuilder)
    def _generate_examples(self, **kwargs):
        model_path = sys.path[0] + "/data/bert/tf_hub/bert_zh_L-12_H-768_A-12/1"
        tokenizer = FullTokenizer(vocab_file=model_path + "/assets/vocab.txt", do_lower_case=True)
        data_path = kwargs.get("data_path")
        data_type = kwargs.get("data_type")
        max_seq_length = int(kwargs.get("max_seq_length"))
        label_map = {}
        for (i, label) in enumerate(list(self.LABEL_DICT.values()), 1):
            label_map[label] = i
        lines = self.readfile(data_path)
        # create examples
        examples = []
        rows = []
        for (i, line) in enumerate(lines):
            if i % 5000 == 0:
                print("Writing example %d of %d" % (i, len(examples)))
            guid = "%s-%s" % (data_type, i)
            textlist = [convert_to_unicode(text) for text in line[0]]
            labellist = [convert_to_unicode(text) for text in line[1]]
            tokens = []
            labels = []
            # print(textlist)
            for idx, word in enumerate(textlist):
                token = tokenizer.tokenize(word)
                # print(token)
                tokens.extend(token)
                label_1 = labellist[idx]
                for m in range(len(token)):
                    if m == 0:
                        labels.append(label_1)
                    else:
                        labels.append("X")
                # print(tokens, labels)
            # tokens = tokenizer.tokenize(example.text)
            if len(tokens) >= max_seq_length - 1:
                tokens = tokens[0:(max_seq_length - 2)]
                labels = labels[0:(max_seq_length - 2)]
            ntokens = []
            segment_ids = []
            label_ids = []
            ntokens.append("[CLS]")
            segment_ids.append(0)
            # append("O") or append("[CLS]") not sure!
            label_ids.append(label_map["[CLS]"])
            for i, token in enumerate(tokens):
                ntokens.append(token)
                segment_ids.append(0)
                label_ids.append(label_map[labels[i]])
            ntokens.append("[SEP]")
            segment_ids.append(0)
            # append("O") or append("[SEP]") not sure!
            label_ids.append(label_map["[SEP]"])
            input_ids = tokenizer.convert_tokens_to_ids(ntokens)
            input_mask = [1] * len(input_ids)
            # label_mask = [1] * len(input_ids)
            while len(input_ids) < max_seq_length:
                input_ids.append(0)
                input_mask.append(0)
                segment_ids.append(0)
                # we don't concerned about it!
                label_ids.append(0)
                ntokens.append("**NULL**")
                # label_mask.append(0)
            # print(len(input_ids))
            assert len(input_ids) == max_seq_length
            assert len(input_mask) == max_seq_length
            assert len(segment_ids) == max_seq_length
            assert len(label_ids) == max_seq_length
            # assert len(label_mask) == max_seq_length
            if i < 5:
                print("*** Example ***")
                print("guid: %s" % (guid))
                print("tokens: %s" % " ".join([printable_text(x) for x in tokens]))
                print("input_ids: %s" % " ".join([str(x) for x in input_ids]))
                print("input_mask: %s" % " ".join([str(x) for x in input_mask]))
                print("segment_ids: %s" % " ".join([str(x) for x in segment_ids]))
                print("label_ids: %s" % " ".join([str(x) for x in label_ids]))
                # print("label_mask: %s" % " ".join([str(x) for x in label_mask]))
            row = {
                "input_ids": input_ids,
                "input_mask": input_mask,
                "segment_ids": segment_ids,
                "target": label_ids
            }
            rows.append(row)
        for curr_row in rows:
            label_val = curr_row.pop(self.TARGET_COLUMN)
            yield {
                "label": label_val,
                "feature": {
                    name: self.FEATURE_DICT[name][1](value)
                    for name, value in curr_row.items()
                }
            }


if __name__ == '__main__':
    dataset_name = ChineseBertNerDatasetBuilder.camelcase_to_snakecase("ChineseBertNerDatasetBuilder")
    print(SuperDatasetBuilder.list_builders())
    batched_datasets = ChineseBertNerDatasetBuilder.load_batched_datasets(dataset_name, config_name="plain_text")
    num_examples = ChineseBertNerDatasetBuilder.get_num_examples(dataset_name, tfds.Split.TRAIN)
    print(f"num_examples={num_examples}")
    batched_dataset_features = ChineseBertNerDatasetBuilder.read_batched_datasets(batched_datasets)
    for features, labels in batched_dataset_features:
        print(f"features={features}, labels={labels}")
    print("end!!")



# -*- coding: utf-8 -*-
import numpy as np
from six import with_metaclass
from abc import ABCMeta, abstractmethod
from py_common_util.common.annotations import PublicAPI


@PublicAPI
class LayerV2(with_metaclass(ABCMeta)):
    """
    自定义layer
    """
    def __init__(self):
        # self.inputs = None
        pass

    def forward(self, *inputs, **kwargs):
        # self.inputs = inputs[0]
        return None

    def call(self, *input, **kwargs):
        """
        layer = XXLayerV2()
        有了__call__方法就可以方便直接调用model(batch_x, parser_args=self.parser_args)方法
        来执行下面forward方法的内容
        :param input:
        :param kwargs:
        :return:
        """
        # print("TfLayerV2#__call__...")
        result = self.forward(*input, **kwargs)
        return result

    def backward(self, grad):
        # 暂时使用后端框架(如tensorflow,keras,pytorch)的自动微分机制
        raise NotImplementedError


@PublicAPI
class NumpyReLU(LayerV2):

    def __init__(self):
        super().__init__()
        self.inputs = None

    def forward(self, *inputs, **kwargs):
        self.inputs = None
        super().forward(*inputs, **kwargs)
        return np.maximum(self.inputs, 0.0)

    def backward(self, grad):
        return self._derivative(self.inputs) * grad

    def _derivative(self, x):
        return x > 0.0




# -*- coding: utf-8 -*-
from skydl.model.default_keras_model import DefaultKerasModel
from skydl.model.train_phase_enum import TrainPhaseEnum
import pandas as pd
import numpy as np
import sys,os
sys.path.append(os.path.abspath('%s/..' % sys.path[0]))


class WideDeepKerasModel(DefaultKerasModel):
    """
    wide&deep model, paper: https://arxiv.org/pdf/1606.07792.pdf
    """
    def adjust_parse_args_value(self):
        super().adjust_parse_args_value()
        self.parser_args.data_path = sys.path[0] + '/../../dataset'
        self.parser_args.use_cuda = True
        self.parser_args.init_from_saver = True
        self.parser_args.train_phase = TrainPhaseEnum.Train.value
        self.parser_args.model_version = '1'
        self.parser_args.epochs = 10
        self.parser_args.batch_size = 128
        self.parser_args.log_interval = 1000

    def load_data(self):
        super().load_data()
        # Get the data: original source is here: https://www.kaggle.com/zynicide/wine-reviews/data
        # URL = "https://storage.googleapis.com/sara-cloud-ml/wine_data.csv"
        # path = tf.keras.utils.get_file(URL.split('/')[-1], URL)
        wine_data = pd.read_csv(self.parser_args.data_path + "/wine_data/wine_data.csv")
        # Shuffle the data
        wine_data = wine_data.sample(frac=1)
        # Print the first 5 rows
        print(wine_data.head())
        # Do some preprocessing to limit the # of wine varities in the dataset
        wine_data = wine_data[pd.notnull(wine_data['country'])]
        wine_data = wine_data[pd.notnull(wine_data['price'])]
        wine_data = wine_data.drop(wine_data.columns[0], axis=1)
        variety_threshold = 500  # Anything that occurs less than this will be removed.
        value_counts = wine_data['variety'].value_counts()
        to_remove = value_counts[value_counts <= variety_threshold].index
        wine_data.replace(to_remove, np.nan, inplace=True)
        wine_data = wine_data[pd.notnull(wine_data['variety'])]
        # Split data into train and test
        train_size = int(len(wine_data) * 0.8)
        print("Train size: %d" % train_size)
        print("Test size: %d" % (len(wine_data) - train_size))
        # Train features
        description_train = wine_data['description'][:train_size]
        variety_train = wine_data['variety'][:train_size]
        # Train labels
        labels_train = wine_data['price'][:train_size]
        # Test features
        description_test = wine_data['description'][train_size:]
        variety_test = wine_data['variety'][train_size:]
        # Test labels
        labels_test = wine_data['price'][train_size:]
        return description_train, variety_train, labels_train, description_test, variety_test, labels_test

    def fit(self):
        if not self.is_training_phase():
            return self
        from sklearn.preprocessing import LabelEncoder
        from tensorflow.python import keras
        description_train, \
        variety_train, \
        labels_train, \
        description_test, \
        variety_test, \
        labels_test = self.load_data()
        # Create a tokenizer to preprocess our text descriptions
        vocab_size = 12000  # This is a hyperparameter, experiment with different values for your dataset
        tokenize = keras.preprocessing.text.Tokenizer(num_words=vocab_size, char_level=False)
        tokenize.fit_on_texts(description_train)  # only fit on train
        # vocab_size = len(tokenize.word_counts.keys())
        # print(vocab_size)
        # Wide feature 1: sparse bag of words (bow) vocab_size vector
        description_bow_train = tokenize.texts_to_matrix(description_train)
        description_bow_test = tokenize.texts_to_matrix(description_test)
        # Wide feature 2: one-hot vector of variety categories
        # Use sklearn utility to convert label strings to numbered index
        encoder = LabelEncoder()
        encoder.fit(variety_train)
        variety_train = encoder.transform(variety_train)
        variety_test = encoder.transform(variety_test)
        num_classes = np.max(variety_train) + 1
        # Convert labels to one hot
        variety_train = keras.utils.to_categorical(variety_train, num_classes)
        variety_test = keras.utils.to_categorical(variety_test, num_classes)
        # Define our wide model with the functional API
        bow_inputs = keras.layers.Input(shape=(vocab_size,))
        variety_inputs = keras.layers.Input(shape=(num_classes,))
        merged_layer = keras.layers.concatenate([bow_inputs, variety_inputs])
        merged_layer = keras.layers.Dense(256, activation='relu')(merged_layer)
        predictions = keras.layers.Dense(1)(merged_layer)
        wide_model = keras.Model(inputs=[bow_inputs, variety_inputs], outputs=predictions)
        wide_model.compile(loss='mse', optimizer='adam', metrics=['accuracy'])
        print(wide_model.summary())
        # Deep model feature: word embeddings of wine descriptions
        train_embed = tokenize.texts_to_sequences(description_train)
        test_embed = tokenize.texts_to_sequences(description_test)
        max_seq_length = 170
        train_embed = keras.preprocessing.sequence.pad_sequences(train_embed, maxlen=max_seq_length, padding="post")
        test_embed = keras.preprocessing.sequence.pad_sequences(test_embed, maxlen=max_seq_length, padding="post")
        # Define our deep model with the Functional API
        deep_inputs = keras.layers.Input(shape=(max_seq_length,))
        embedding = keras.layers.Embedding(vocab_size, 8, input_length=max_seq_length)(deep_inputs)
        embedding = keras.layers.Flatten()(embedding)
        embed_out = keras.layers.Dense(1)(embedding)
        deep_model = keras.Model(inputs=deep_inputs, outputs=embed_out)
        print(deep_model.summary())
        deep_model.compile(loss='mse',
                           optimizer='adam',
                           metrics=['accuracy'])
        # Combine wide and deep into one model
        merged_out = keras.layers.concatenate([wide_model.output, deep_model.output])
        merged_out = keras.layers.Dense(1)(merged_out)
        combined_model = keras.Model(wide_model.input + [deep_model.input], merged_out)
        print(combined_model.summary())
        combined_model.compile(loss='mse',
                               optimizer='adam',
                               metrics=['accuracy'])
        # Run training
        combined_model.fit([description_bow_train, variety_train] + [train_embed], labels_train, epochs=self.parser_args.epochs, batch_size=self.parser_args.batch_size)
        combined_model.evaluate([description_bow_test, variety_test] + [test_embed], labels_test, batch_size=self.parser_args.batch_size)
        # Generate predictions
        predictions = combined_model.predict([description_bow_test, variety_test] + [test_embed])
        # Compare predictions with actual values for the first few items in our test dataset
        num_predictions = 40
        diff = 0
        for i in range(num_predictions):
            val = predictions[i]
            print(description_test.iloc[i])
            print('Predicted: ', val[0], 'Actual: ', labels_test.iloc[i], '\n')
            diff += abs(val[0] - labels_test.iloc[i])
        # Compare the average difference between actual price and the model's predicted price
        print('Average prediction difference: ', diff / num_predictions)
        return self

    def evaluate(self, *args, **kwargs):
        return self

    def serving(self):
        return self




# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Trainer network for BERT-style models."""

from __future__ import absolute_import
from __future__ import division
# from __future__ import google_type_annotations
from __future__ import print_function

import copy
import tensorflow as tf

from skydl.models_zoo.bert.official.nlp.modeling import networks


@tf.keras.utils.register_keras_serializable(package='Text')
class BertPretrainer(tf.keras.Model):
  """BERT network training model.

  This is an implementation of the network structure surrounding a transformer
  encoder as described in "BERT: Pre-training of Deep Bidirectional Transformers
  for Language Understanding" (https://arxiv.org/abs/1810.04805).

  The BertTrainer allows a user to pass in a transformer stack, and instantiates
  the masked language model and classification networks that are used to create
  the training objectives.

  Attributes:
    network: A transformer network. This network should output a sequence output
      and a classification output. Furthermore, it should expose its embedding
      table via a "get_embedding_table" method.
    num_classes: Number of classes to predict from the classification network.
    num_token_predictions: Number of tokens to predict from the masked LM.
    activation: The activation (if any) to use in the masked LM and
      classification networks. If None, no activation will be used.
    initializer: The initializer (if any) to use in the masked LM and
      classification networks. Defaults to a Glorot uniform initializer.
    output: The output style for this network. Can be either 'logits' or
      'predictions'.
  """

  def __init__(self,
               network,
               num_classes,
               num_token_predictions,
               activation=None,
               output_activation=None,
               initializer='glorot_uniform',
               output='logits',
               **kwargs):
    self._self_setattr_tracking = False
    self._config = {
        'network': network,
        'num_classes': num_classes,
        'num_token_predictions': num_token_predictions,
        'activation': activation,
        'output_activation': output_activation,
        'initializer': initializer,
        'output': output,
    }

    # We want to use the inputs of the passed network as the inputs to this
    # Model. To do this, we need to keep a copy of the network inputs for use
    # when we construct the Model object at the end of init. (We keep a copy
    # because we'll be adding another tensor to the copy later.)
    network_inputs = network.inputs
    inputs = copy.copy(network_inputs)

    # Because we have a copy of inputs to create this Model object, we can
    # invoke the Network object with its own input tensors to start the Model.
    # Note that, because of how deferred construction happens, we can't use
    # the copy of the list here - by the time the network is invoked, the list
    # object contains the additional input added below.
    sequence_output, cls_output = network(network_inputs)

    sequence_output_length = sequence_output.shape.as_list()[1]
    if sequence_output_length < num_token_predictions:
      raise ValueError(
          "The passed network's output length is %s, which is less than the "
          'requested num_token_predictions %s.' %
          (sequence_output_length, num_token_predictions))

    masked_lm_positions = tf.keras.layers.Input(
        shape=(num_token_predictions,),
        name='masked_lm_positions',
        dtype=tf.int32)
    inputs.append(masked_lm_positions)

    self.masked_lm = networks.MaskedLM(
        num_predictions=num_token_predictions,
        input_width=sequence_output.shape[-1],
        source_network=network,
        activation=activation,
        initializer=initializer,
        output=output,
        name='masked_lm')
    lm_outputs = self.masked_lm([sequence_output, masked_lm_positions])

    self.classification = networks.Classification(
        input_width=cls_output.shape[-1],
        num_classes=num_classes,
        initializer=initializer,
        output=output,
        name='classification')
    sentence_outputs = self.classification(cls_output)

    super(BertPretrainer, self).__init__(
        inputs=inputs, outputs=[lm_outputs, sentence_outputs], **kwargs)

  def get_config(self):
    return self._config

  @classmethod
  def from_config(cls, config, custom_objects=None):
    return cls(**config)



class Config:
    STREAMING_JOB_NAME = "streaming.job.name"
    STREAMING_OP_NAME = "streaming.op_name"
    TASK_JOB_ID = "streaming.task_job_id"
    STREAMING_WORKER_NAME = "streaming.worker_name"
    # channel
    CHANNEL_TYPE = "channel_type"
    MEMORY_CHANNEL = "memory_channel"
    NATIVE_CHANNEL = "native_channel"
    CHANNEL_SIZE = "channel_size"
    CHANNEL_SIZE_DEFAULT = 10**8
    IS_RECREATE = "streaming.is_recreate"
    # return from StreamingReader.getBundle if only empty message read in this
    # interval.
    TIMER_INTERVAL_MS = "timer_interval_ms"

    STREAMING_RING_BUFFER_CAPACITY = "streaming.ring_buffer_capacity"
    # write an empty message if there is no data to be written in this
    # interval.
    STREAMING_EMPTY_MESSAGE_INTERVAL = "streaming.empty_message_interval"

    # operator type
    OPERATOR_TYPE = "operator_type"



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import sys
import uuid

import networkx as nx

from skydl.ray.experimental.streaming.communication import DataChannel, DataInput
from skydl.ray.experimental.streaming.communication import DataOutput, QueueConfig
from skydl.ray.experimental.streaming.operator import Operator, OpType
from skydl.ray.experimental.streaming.operator import PScheme, PStrategy
import skydl.ray.experimental.streaming.operator_instance as operator_instance

logger = logging.getLogger(__name__)
logger.setLevel("INFO")


# Generates UUIDs
def _generate_uuid():
    return uuid.uuid4()


# Rolling sum's logic
def _sum(value_1, value_2):
    return value_1 + value_2


# Partitioning strategies that require all-to-all instance communication
all_to_all_strategies = [
    PStrategy.Shuffle, PStrategy.ShuffleByKey, PStrategy.Broadcast,
    PStrategy.RoundRobin
]


# Environment configuration
class Config(object):
    """Environment configuration.

    This class includes all information about the configuration of the
    streaming environment.

    Attributes:
         queue_config (QueueConfig): Batched Queue configuration
         (see: communication.py)
         A batched queue configuration includes the max queue size,
         the size of each batch (in number of elements), the batch flush
         timeout, and the number of batches to prefetch from plasma
         parallelism (int): The number of isntances (actors) for each logical
         dataflow operator (default: 1)
    """

    def __init__(self, parallelism=1):
        self.queue_config = QueueConfig()
        self.parallelism = parallelism
        # ...


# The execution environment for a streaming job
class Environment(object):
    """A streaming environment.

    This class is responsible for constructing the logical and the
    physical dataflow.

    Attributes:
         logical_topo (DiGraph): The user-defined logical topology in
         NetworkX DiGRaph format.
         (See: https://networkx.github.io)
         physical_topo (DiGraph): The physical topology in NetworkX
         DiGRaph format. The physical dataflow is constructed by the
         environment based on logical_topo.
         operators (dict): A mapping from operator ids to operator metadata
         (See: Operator in operator.py).
         config (Config): The environment's configuration.
         topo_cleaned (bool): A flag that indicates whether the logical
         topology is garbage collected (True) or not (False).
         actor_handles (list): A list of all Ray actor handles that execute
         the streaming dataflow.
    """

    def __init__(self, config=Config()):
        self.logical_topo = nx.DiGraph()  # DAG
        self.physical_topo = nx.DiGraph()  # DAG
        self.operators = {}  # operator id --> operator object
        self.config = config  # Environment's configuration
        self.topo_cleaned = False
        # Handles to all actors in the physical dataflow
        self.actor_handles = []

    # Constructs and deploys a Ray actor of a specific type
    # TODO (john): Actor placement information should be specified in
    # the environment's configuration
    def __generate_actor(self, instance_id, operator, input, output):
        """Generates an actor that will execute a particular instance of
        the logical operator

        Attributes:
            instance_id (UUID): The id of the instance the actor will execute.
            operator (Operator): The metadata of the logical operator.
            input (DataInput): The input gate that manages input channels of
            the instance (see: DataInput in communication.py).
            input (DataOutput): The output gate that manages output channels
            of the instance (see: DataOutput in communication.py).
        """
        actor_id = (operator.id, instance_id)
        # Record the physical dataflow graph (for debugging purposes)
        self.__add_channel(actor_id, input, output)
        # Select actor to construct
        if operator.type == OpType.Source:
            source = operator_instance.Source.remote(actor_id, operator, input,
                                                     output)
            source.register_handle.remote(source)
            return source.start.remote()
        elif operator.type == OpType.Map:
            map = operator_instance.Map.remote(actor_id, operator, input,
                                               output)
            map.register_handle.remote(map)
            return map.start.remote()
        elif operator.type == OpType.FlatMap:
            flatmap = operator_instance.FlatMap.remote(actor_id, operator,
                                                       input, output)
            flatmap.register_handle.remote(flatmap)
            return flatmap.start.remote()
        elif operator.type == OpType.Filter:
            filter = operator_instance.Filter.remote(actor_id, operator, input,
                                                     output)
            filter.register_handle.remote(filter)
            return filter.start.remote()
        elif operator.type == OpType.Reduce:
            reduce = operator_instance.Reduce.remote(actor_id, operator, input,
                                                     output)
            reduce.register_handle.remote(reduce)
            return reduce.start.remote()
        elif operator.type == OpType.TimeWindow:
            pass
        elif operator.type == OpType.KeyBy:
            keyby = operator_instance.KeyBy.remote(actor_id, operator, input,
                                                   output)
            keyby.register_handle.remote(keyby)
            return keyby.start.remote()
        elif operator.type == OpType.Sum:
            sum = operator_instance.Reduce.remote(actor_id, operator, input,
                                                  output)
            # Register target handle at state actor
            state_actor = operator.state_actor
            if state_actor is not None:
                state_actor.register_target.remote(sum)
            # Register own handle
            sum.register_handle.remote(sum)
            return sum.start.remote()
        elif operator.type == OpType.Sink:
            pass
        elif operator.type == OpType.Inspect:
            inspect = operator_instance.Inspect.remote(actor_id, operator,
                                                       input, output)
            inspect.register_handle.remote(inspect)
            return inspect.start.remote()
        elif operator.type == OpType.ReadTextFile:
            # TODO (john): Colocate the source with the input file
            read = operator_instance.ReadTextFile.remote(
                actor_id, operator, input, output)
            read.register_handle.remote(read)
            return read.start.remote()
        else:  # TODO (john): Add support for other types of operators
            sys.exit("Unrecognized or unsupported {} operator type.".format(
                operator.type))

    # Constructs and deploys a Ray actor for each instance of
    # the given operator
    def __generate_actors(self, operator, upstream_channels,
                          downstream_channels):
        """Generates one actor for each instance of the given logical
        operator.

        Attributes:
            operator (Operator): The logical operator metadata.
            upstream_channels (list): A list of all upstream channels for
            all instances of the operator.
            downstream_channels (list): A list of all downstream channels
            for all instances of the operator.
        """
        num_instances = operator.num_instances
        logger.info("Generating {} actors of type {}...".format(
            num_instances, operator.type))
        in_channels = upstream_channels.pop(
            operator.id) if upstream_channels else []
        handles = []
        for i in range(num_instances):
            # Collect input and output channels for the particular instance
            ip = [
                channel for channel in in_channels
                if channel.dst_instance_id == i
            ] if in_channels else []
            op = [
                channel for channels_list in downstream_channels.values()
                for channel in channels_list if channel.src_instance_id == i
            ]
            log = "Constructed {} input and {} output channels "
            log += "for the {}-th instance of the {} operator."
            logger.debug(log.format(len(ip), len(op), i, operator.type))
            input_gate = DataInput(ip)
            output_gate = DataOutput(op, operator.partitioning_strategies)
            handle = self.__generate_actor(i, operator, input_gate,
                                           output_gate)
            if handle:
                handles.append(handle)
        return handles

    # Adds a channel/edge to the physical dataflow graph
    def __add_channel(self, actor_id, input, output):
        for dest_actor_id in output._destination_actor_ids():
            self.physical_topo.add_edge(actor_id, dest_actor_id)

    # Generates all required data channels between an operator
    # and its downstream operators
    def _generate_channels(self, operator):
        """Generates all output data channels
        (see: DataChannel in communication.py) for all instances of
        the given logical operator.

        The function constructs one data channel for each pair of
        communicating operator instances (instance_1,instance_2),
        where instance_1 is an instance of the given operator and instance_2
        is an instance of a direct downstream operator.

        The number of total channels generated depends on the partitioning
        strategy specified by the user.
        """
        channels = {}  # destination operator id -> channels
        strategies = operator.partitioning_strategies
        for dst_operator, p_scheme in strategies.items():
            num_dest_instances = self.operators[dst_operator].num_instances
            entry = channels.setdefault(dst_operator, [])
            if p_scheme.strategy == PStrategy.Forward:
                for i in range(operator.num_instances):
                    # ID of destination instance to connect
                    id = i % num_dest_instances
                    channel = DataChannel(self, operator.id, dst_operator, i,
                                          id)
                    entry.append(channel)
            elif p_scheme.strategy in all_to_all_strategies:
                for i in range(operator.num_instances):
                    for j in range(num_dest_instances):
                        channel = DataChannel(self, operator.id, dst_operator,
                                              i, j)
                        entry.append(channel)
            else:
                # TODO (john): Add support for other partitioning strategies
                sys.exit("Unrecognized or unsupported partitioning strategy.")
        return channels

    # An edge denotes a flow of data between logical operators
    # and may correspond to multiple data channels in the physical dataflow
    def _add_edge(self, source, destination):
        self.logical_topo.add_edge(source, destination)

    # Cleans the logical dataflow graph to construct and
    # deploy the physical dataflow
    def _collect_garbage(self):
        if self.topo_cleaned is True:
            return
        for node in self.logical_topo:
            self.operators[node]._clean()
        self.topo_cleaned = True

    # Sets the level of parallelism for a registered operator
    # Overwrites the environment parallelism (if set)
    def _set_parallelism(self, operator_id, level_of_parallelism):
        self.operators[operator_id].num_instances = level_of_parallelism

    # Sets the same level of parallelism for all operators in the environment
    def set_parallelism(self, parallelism):
        self.config.parallelism = parallelism

    # Sets batched queue configuration for the environment
    def set_queue_config(self, queue_config):
        self.config.queue_config = queue_config

    # Creates and registers a user-defined data source
    # TODO (john): There should be different types of sources, e.g. sources
    # reading from Kafka, text files, etc.
    # TODO (john): Handle case where environment parallelism is set
    def source(self, source):
        source_id = _generate_uuid()
        source_stream = DataStream(self, source_id)
        self.operators[source_id] = Operator(
            source_id, OpType.Source, "Source", other=source)
        return source_stream

    # Creates and registers a new data source that reads a
    # text file line by line
    # TODO (john): There should be different types of sources,
    # e.g. sources reading from Kafka, text files, etc.
    # TODO (john): Handle case where environment parallelism is set
    def read_text_file(self, filepath):
        source_id = _generate_uuid()
        source_stream = DataStream(self, source_id)
        self.operators[source_id] = Operator(
            source_id, OpType.ReadTextFile, "Read Text File", other=filepath)
        return source_stream

    # Constructs and deploys the physical dataflow
    def execute(self):
        """Deploys and executes the physical dataflow."""
        self._collect_garbage()  # Make sure everything is clean
        # TODO (john): Check if dataflow has any 'logical inconsistencies'
        # For example, if there is a forward partitioning strategy but
        # the number of downstream instances is larger than the number of
        # upstream instances, some of the downstream instances will not be
        # used at all

        # Each operator instance is implemented as a Ray actor
        # Actors are deployed in topological order, as we traverse the
        # logical dataflow from sources to sinks. At each step, data
        # producers wait for acknowledge from consumers before starting
        # generating data.
        upstream_channels = {}
        for node in nx.topological_sort(self.logical_topo):
            operator = self.operators[node]
            # Generate downstream data channels
            downstream_channels = self._generate_channels(operator)
            # Instantiate Ray actors
            handles = self.__generate_actors(operator, upstream_channels,
                                             downstream_channels)
            if handles:
                self.actor_handles.extend(handles)
            upstream_channels.update(downstream_channels)
        logger.debug("Running...")
        return self.actor_handles

    # Prints the logical dataflow graph
    def print_logical_graph(self):
        self._collect_garbage()
        logger.info("==================================")
        logger.info("======Logical Dataflow Graph======")
        logger.info("==================================")
        # Print operators in topological order
        for node in nx.topological_sort(self.logical_topo):
            downstream_neighbors = list(self.logical_topo.neighbors(node))
            logger.info("======Current Operator======")
            operator = self.operators[node]
            operator.print()
            logger.info("======Downstream Operators======")
            if len(downstream_neighbors) == 0:
                logger.info("None\n")
            for downstream_node in downstream_neighbors:
                self.operators[downstream_node].print()

    # Prints the physical dataflow graph
    def print_physical_graph(self):
        logger.info("===================================")
        logger.info("======Physical Dataflow Graph======")
        logger.info("===================================")
        # Print all data channels between operator instances
        log = "(Source Operator ID,Source Operator Name,Source Instance ID)"
        log += " --> "
        log += "(Destination Operator ID,Destination Operator Name,"
        log += "Destination Instance ID)"
        logger.info(log)
        for src_actor_id, dst_actor_id in self.physical_topo.edges:
            src_operator_id, src_instance_id = src_actor_id
            dst_operator_id, dst_instance_id = dst_actor_id
            logger.info("({},{},{}) --> ({},{},{})".format(
                src_operator_id, self.operators[src_operator_id].name,
                src_instance_id, dst_operator_id,
                self.operators[dst_operator_id].name, dst_instance_id))


# TODO (john): We also need KeyedDataStream and WindowedDataStream as
# subclasses of DataStream to prevent ill-defined logical dataflows


# A DataStream corresponds to an edge in the logical dataflow
class DataStream(object):
    """A data stream.

    This class contains all information about a logical stream, i.e. an edge
    in the logical topology. It is the main class exposed to the user.

    Attributes:
         id (UUID): The id of the stream
         env (Environment): The environment the stream belongs to.
         src_operator_id (UUID): The id of the source operator of the stream.
         dst_operator_id (UUID): The id of the destination operator of the
         stream.
         is_partitioned (bool): Denotes if there is a partitioning strategy
         (e.g. shuffle) for the stream or not (default stategy: Forward).
    """

    def __init__(self,
                 environment,
                 source_id=None,
                 dest_id=None,
                 is_partitioned=False):
        self.id = _generate_uuid()
        self.env = environment
        self.src_operator_id = source_id
        self.dst_operator_id = dest_id
        # True if a partitioning strategy for this stream exists,
        # false otherwise
        self.is_partitioned = is_partitioned

    # Generates a new stream after a data transformation is applied
    def __expand(self):
        stream = DataStream(self.env)
        assert (self.dst_operator_id is not None)
        stream.src_operator_id = self.dst_operator_id
        stream.dst_operator_id = None
        return stream

    # Assigns the partitioning strategy to a new 'open-ended' stream
    # and returns the stream. At this point, the partitioning strategy
    # is not associated with any destination operator. We expect this to
    # be done later, as we continue assembling the dataflow graph
    def __partition(self, strategy, partition_fn=None):
        scheme = PScheme(strategy, partition_fn)
        source_operator = self.env.operators[self.src_operator_id]
        new_stream = DataStream(
            self.env, source_id=source_operator.id, is_partitioned=True)
        source_operator._set_partition_strategy(new_stream.id, scheme)
        return new_stream

    # Registers the operator to the environment and returns a new
    # 'open-ended' stream. The registered operator serves as the destination
    # of the previously 'open' stream
    def __register(self, operator):
        """Registers the given logical operator to the environment and
        connects it to its upstream operator (if any).

        A call to this function adds a new edge to the logical topology.

        Attributes:
             operator (Operator): The metadata of the logical operator.
        """
        self.env.operators[operator.id] = operator
        self.dst_operator_id = operator.id
        logger.debug("Adding new dataflow edge ({},{}) --> ({},{})".format(
            self.src_operator_id,
            self.env.operators[self.src_operator_id].name,
            self.dst_operator_id,
            self.env.operators[self.dst_operator_id].name))
        # Update logical dataflow graphs
        self.env._add_edge(self.src_operator_id, self.dst_operator_id)
        # Keep track of the partitioning strategy and the destination operator
        src_operator = self.env.operators[self.src_operator_id]
        if self.is_partitioned is True:
            partitioning, _ = src_operator._get_partition_strategy(self.id)
            src_operator._set_partition_strategy(_generate_uuid(),
                                                 partitioning, operator.id)
        elif src_operator.type == OpType.KeyBy:
            # Set the output partitioning strategy to shuffle by key
            partitioning = PScheme(PStrategy.ShuffleByKey)
            src_operator._set_partition_strategy(_generate_uuid(),
                                                 partitioning, operator.id)
        else:  # No partitioning strategy has been defined - set default
            partitioning = PScheme(PStrategy.Forward)
            src_operator._set_partition_strategy(_generate_uuid(),
                                                 partitioning, operator.id)
        return self.__expand()

    # Sets the level of parallelism for an operator, i.e. its total
    # number of instances. Each operator instance corresponds to an actor
    # in the physical dataflow
    def set_parallelism(self, num_instances):
        """Sets the number of instances for the source operator of the stream.

        Attributes:
             num_instances (int): The level of parallelism for the source
             operator of the stream.
        """
        assert (num_instances > 0)
        self.env._set_parallelism(self.src_operator_id, num_instances)
        return self

    # Stream Partitioning Strategies #
    # TODO (john): Currently, only forward (default), shuffle,
    # and broadcast are supported

    # Hash-based record shuffling
    def shuffle(self):
        """Registers a shuffling partitioning strategy for the stream."""
        return self.__partition(PStrategy.Shuffle)

    # Broadcasts each record to all downstream instances
    def broadcast(self):
        """Registers a broadcast partitioning strategy for the stream."""
        return self.__partition(PStrategy.Broadcast)

    # Rescales load to downstream instances
    def rescale(self):
        """Registers a rescale partitioning strategy for the stream.

        Same as Flink's rescale (see: https://ci.apache.org/projects/flink/
        flink-docs-stable/dev/stream/operators/#physical-partitioning).
        """
        return self.__partition(PStrategy.Rescale)

    # Round-robin partitioning
    def round_robin(self):
        """Registers a round-robin partitioning strategy for the stream."""
        return self.__partition(PStrategy.RoundRobin)

    # User-defined partitioning
    def partition(self, partition_fn):
        """Registers a user-defined partitioning strategy for the stream.

        Attributes:
             partition_fn (function): The user-defined partitioning function.
        """
        return self.__partition(PStrategy.Custom, partition_fn)

    #   Data Trasnformations   #
    # TODO (john): Expand set of supported operators.
    # TODO (john): To support event-time windows we need a mechanism for
    # generating and processing watermarks

    # Registers map operator to the environment
    def map(self, map_fn, name="Map"):
        """Applies a map operator to the stream.

        Attributes:
             map_fn (function): The user-defined logic of the map.
        """
        op = Operator(
            _generate_uuid(),
            OpType.Map,
            name,
            map_fn,
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # Registers flatmap operator to the environment
    def flat_map(self, flatmap_fn):
        """Applies a flatmap operator to the stream.

        Attributes:
             flatmap_fn (function): The user-defined logic of the flatmap
             (e.g. split()).
        """
        op = Operator(
            _generate_uuid(),
            OpType.FlatMap,
            "FlatMap",
            flatmap_fn,
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # Registers keyBy operator to the environment
    # TODO (john): This should returned a KeyedDataStream
    def key_by(self, key_selector):
        """Applies a key_by operator to the stream.

        Attributes:
             key_attribute_index (int): The index of the key attributed
             (assuming tuple records).
        """
        op = Operator(
            _generate_uuid(),
            OpType.KeyBy,
            "KeyBy",
            other=key_selector,
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # Registers Reduce operator to the environment
    def reduce(self, reduce_fn):
        """Applies a rolling sum operator to the stream.

        Attributes:
             sum_attribute_index (int): The index of the attribute to sum
             (assuming tuple records).
        """
        op = Operator(
            _generate_uuid(),
            OpType.Reduce,
            "Sum",
            reduce_fn,
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # Registers Sum operator to the environment
    def sum(self, attribute_selector, state_keeper=None):
        """Applies a rolling sum operator to the stream.

        Attributes:
             sum_attribute_index (int): The index of the attribute to sum
             (assuming tuple records).
        """
        op = Operator(
            _generate_uuid(),
            OpType.Sum,
            "Sum",
            _sum,
            other=attribute_selector,
            state_actor=state_keeper,
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # Registers window operator to the environment.
    # This is a system time window
    # TODO (john): This should return a WindowedDataStream
    def time_window(self, window_width_ms):
        """Applies a system time window to the stream.

        Attributes:
             window_width_ms (int): The length of the window in ms.
        """
        op = Operator(
            _generate_uuid(),
            OpType.TimeWindow,
            "TimeWindow",
            num_instances=self.env.config.parallelism,
            other=window_width_ms)
        return self.__register(op)

    # Registers filter operator to the environment
    def filter(self, filter_fn):
        """Applies a filter to the stream.

        Attributes:
             filter_fn (function): The user-defined filter function.
        """
        op = Operator(
            _generate_uuid(),
            OpType.Filter,
            "Filter",
            filter_fn,
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # TODO (john): Registers window join operator to the environment
    def window_join(self, other_stream, join_attribute, window_width):
        op = Operator(
            _generate_uuid(),
            OpType.WindowJoin,
            "WindowJoin",
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # Registers inspect operator to the environment
    def inspect(self, inspect_logic):
        """Inspects the content of the stream.

        Attributes:
             inspect_logic (function): The user-defined inspect function.
        """
        op = Operator(
            _generate_uuid(),
            OpType.Inspect,
            "Inspect",
            inspect_logic,
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # Registers sink operator to the environment
    # TODO (john): A sink now just drops records but it should be able to
    # export data to other systems
    def sink(self):
        """Closes the stream with a sink operator."""
        op = Operator(
            _generate_uuid(),
            OpType.Sink,
            "Sink",
            num_instances=self.env.config.parallelism)
        return self.__register(op)



# -*- coding: utf-8 -*-
import tensorflow as tf
# 这样import keras可以查看keras源代码
from tensorflow.python import keras


class KerasModelMixin(keras.Sequential):
    """
    参考: Training and Serving ML models with tf.keras  https://medium.com/tensorflow/training-and-serving-ml-models-with-tf-keras-fd975cc0fa27
    https://github.com/tensorflow/docs/blob/master/site/en/r2/tutorials/keras/basic_classification.ipynb
    https://www.tensorflow.org/guide/keras?hl=zh_cn
    注意：keras model subclass必须继承于超类keras.Sequential才能成功执行model_to_estimator()
    """
    @property
    def logits(self):
        return self._logits

    def __init__(self, layers=[], keras_inputs=None, keras_outputs=None):
        super().__init__()
        self._logits = None
        self._keras_inputs = keras_inputs
        self._keras_outputs = keras_outputs
        if keras_inputs is None or keras_outputs is None:
            self.add(layers)
        else:
            # TODO self._init_graph_network(inputs=keras_inputs, outputs=keras_outputs)
            pass

    def add(self, layers=[]):
        """add layer list"""
        self._layers += layers

    def call_layers(self, _x:tf.Tensor=None, name="prediction"):
        """_x: input layer tensor"""
        if self.layers is None or len(self.layers) < 1:
            return tf.constant(None)
        perdiction = self.layers[0](_x)
        for layer in self.layers[1:]:
            perdiction = layer(perdiction)
        self._logits = perdiction
        return self._logits

    def get_static_logits_output(self):
        if self._keras_outputs:
            return self._outputs
        else:
            last_layer = self._layers[0]
            for layer in self._layers[1:]:
                last_layer = layer(last_layer)
            return last_layer

    def get_static_first_input(self):
        if self._keras_inputs:
            return self._inputs
        else:
            return self._layers[0]

    def call(self, _x: tf.Tensor=None, training: bool=True) -> tf.Tensor:
        """Run the keras model layers."""
        # super().call(_x, training)
        return self.call_layers(_x)




import argparse
import logging
import time

import ray
import wikipedia
from ray.streaming import StreamingContext
from ray.streaming.config import Config

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

parser = argparse.ArgumentParser()
parser.add_argument(
    "--titles-file",
    required=True,
    help="the file containing the wikipedia titles to lookup")


# A custom data source that reads articles from wikipedia
# Custom data sources need to implement a get_next() method
# that returns the next data element, in this case sentences
class Wikipedia:
    def __init__(self, title_file):
        # Titles in this file will be as queries
        self.title_file = title_file
        self.title_reader = iter(list(open(self.title_file, "r").readlines()))
        self.done = False
        self.article_done = True
        self.sentences = iter([])

    # Returns next sentence from a wikipedia article
    def get_next(self):
        if self.done:
            return None  # Source exhausted
        while True:
            if self.article_done:
                try:  # Try next title
                    next_title = next(self.title_reader)
                except StopIteration:
                    self.done = True  # Source exhausted
                    return None
                # Get next article
                logger.debug("Next article: {}".format(next_title))
                article = wikipedia.page(next_title).content
                # Split article in sentences
                self.sentences = iter(article.split("."))
                self.article_done = False
            try:  # Try next sentence
                sentence = next(self.sentences)
                logger.debug("Next sentence: {}".format(sentence))
                return sentence
            except StopIteration:
                self.article_done = True


# Splits input line into words and
# outputs records of the form (word,1)
def splitter(line):
    print(line)
    return [(word, 1) for word in line.split()]


if __name__ == "__main__":
    # Get program parameters
    args = parser.parse_args()
    titles_file = str(args.titles_file)

    ray.init(address="auto", load_code_from_local=True, include_java=False)
    print("ray inited...")
    ctx = StreamingContext.Builder() \
        .option(Config.CHANNEL_TYPE, Config.NATIVE_CHANNEL) \
        .build()
    # A Ray streaming environment with the default configuration
    ctx.set_parallelism(1)  # Each operator will be executed by two actors
    print("2222")
    # Reads articles from wikipedia, splits them in words,
    # shuffles words, and counts the occurrences of each word.
    # stream = ctx.source(Wikipedia(titles_file)) \
    stream = ctx.read_text_file(titles_file) \
        .flat_map(splitter) \
        .key_by(lambda x: x[0]) \
        .reduce(lambda old_value, new_value:
                (old_value[0], old_value[1] + new_value[1])) \
        .sink(print)
    print(".......")
    start = time.time()
    ctx.execute("wordcount_v3")
    end = time.time()
    logger.info("Elapsed time: {} secs".format(end - start))


# -*- coding: utf-8 -*-
import tensorflow as tf
from tensorflow import keras


class TfKerasLayerMixin(keras.layers.Layer):

    def __init__(self, units=None, activation=None, **kwargs):
        """
        :param units: e.g. 10
        :param activation: e.g. python function or str: "softmax"|"relu" or wrapped by @tf.function
        :param kwargs: other args
        """
        super().__init__()
        self.activation = keras.layers.Activation(activation)
        self.units = units

    def build(self, input_shape):
        self.w = self.add_weight("W",
                                 shape=(input_shape[-1], self.units),
                                 # initializer='random_normal',
                                 trainable=True)
        self.b = self.add_weight("B",
                                 shape=(self.units,),
                                 # initializer='random_normal',
                                 trainable=True)

    def call(self, inputs):
        return self.activation(tf.matmul(inputs, self.w) + self.b)





# coding: utf-8 or # -*- coding: utf-8 -*-
"""
参考：OpenAI Gym for Trading With DQN OpenAI Baseline  https://github.com/AdrianP-/gym_trading
OpenAI Baseline: https://github.com/openai/baselines
Gym_trading by Peter Henry: https://github.com/Henry-bee/gym_trading/
Also, good thougths in Trading and Q learning: https://github.com/savourylie/Stock-Price-Forecaster
https://github.com/Prediction-Machines/Trading-Gym
https://github.com/Prediction-Machines/Trading-Brain
https://github.com/Kismuz/btgym
"""
# import matplotlib.dates as mdates
# import matplotlib.finance as mf
import matplotlib.pyplot as plt
from .super_env import SuperEnv
from .trading.portfolio import Portfolio
import pandas as pd


class TradingEnv(SuperEnv):

    def __init__(self):
        super().__init__()

    def do_init(self, q_size=4, gamma=0.8, states=[], observation=[], low_box=0, high_box=0, train=True):
        super().do_init(q_size=q_size, gamma=gamma, states=states, observation=observation, low_box=low_box, high_box=high_box, train=train)
        self.portfolio = Portfolio(prices=states[['Open', 'Close']], trade_period=5000, train_end_index=observation.shape[0])

    def reset(self):
        super().reset()
        if self.train:
            obs = self.observation.values[0]
            self.current_index = 1
            self._end = self.observation.shape[0] - 1
        else:
            obs = self.observation.values[0]
            self.current_index = 1
            self._end = self.observation.shape[0] - 1
        self._data = self.states.iloc[self.current_index:self._end + 1]
        self.portfolio._reset(self.train)
        return obs

    def step(self, action):
        super().step(action)
        # Return the observation, done, reward from Simulator and Portfolio
        reward, info = self.portfolio._step(action)
        # obs, done = self.sim._step(self.portfolio.open_trade, self.portfolio.curr_trade["Trade Duration"])
        if self.portfolio.open_trade:
            obs = self.observation.values[self.current_index] + [self.portfolio.open_trade] + [self.portfolio.curr_trade["Trade Duration"]]
        else:
            obs = self.observation.values[self.current_index]
        self.current_index += 1
        done = self.current_index > self._end
        return obs, reward, done, info

    def render(self, mode='human'):
        super().render(mode)

    def seed(self, seed=None):
        return super().seed(seed)

    def generate_summary_stats(self):
        print("SUMMARY STATISTICS")
        journal = pd.DataFrame(self.portfolio.journal)
        print("Total Trades Taken: ", journal.shape[0])
        print("Total Reward: ", journal['Profit'].sum())
        print("Average Reward per Trade: ", journal['Profit'].sum() / journal['Profit'].count())
        print("Win Ratio: %s %%" % (((journal.loc[journal['Profit'] > 0, 'Profit'].count()) / journal.shape[0]) * 100))

        fig, ax = plt.subplots(figsize=(40, 10))

        data = self._data
        # Get a OHLC list with tuples (dates, Open, High, Low, Close)
        # ohlc = list(zip(mdates.date2num(data.index.to_pydatetime()), data.Open.tolist(), data.High.tolist(), data.Low.tolist(), data.Close.tolist()))
        ohlc = list(zip(list(range(data.shape[0])), data.Open.tolist(), data.High.tolist(),
                        data.Low.tolist(), data.Close.tolist()))
        # Filter out buy and sell orders for plotting
        buys = journal.loc[journal.Type == 'BUY', :]
        sells = journal.loc[journal.Type == 'SELL', :]

        # #Plotting functions
        # mf.candlestick_ohlc(ax, ohlc, width=0.02, colorup='green', colordown='red')
        ax.plot(buys['Entry Time'], buys['Entry Price'] - 0.001, 'b^', alpha=1.0)
        ax.plot(sells['Entry Time'], sells['Entry Price'] + 0.001, 'rv', alpha=1.0)

        print("-----save plt to png: trading_env.png")
        plt.savefig("trading_env.png")
        # plt.show()

        import pprint
        pp = pprint.PrettyPrinter(indent=2)
        pp.pprint(self.portfolio.journal)



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Flags for managing compute devices. Currently only contains TPU flags."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import flags
import tensorflow as tf

from skydl.models_zoo.bert.official.utils.flags._conventions import help_wrap


def require_cloud_storage(flag_names):
  """Register a validator to check directory flags.
  Args:
    flag_names: An iterable of strings containing the names of flags to be
      checked.
  """
  msg = "TPU requires GCS path for {}".format(", ".join(flag_names))
  @flags.multi_flags_validator(["tpu"] + flag_names, message=msg)
  def _path_check(flag_values):  # pylint: disable=missing-docstring
    if flag_values["tpu"] is None:
      return True

    valid_flags = True
    for key in flag_names:
      if not flag_values[key].startswith("gs://"):
        tf.compat.v1.logging.error("{} must be a GCS path.".format(key))
        valid_flags = False

    return valid_flags


def define_device(tpu=True):
  """Register device specific flags.
  Args:
    tpu: Create flags to specify TPU operation.
  Returns:
    A list of flags for core.py to marks as key flags.
  """

  key_flags = []

  if tpu:
    flags.DEFINE_string(
        name="tpu", default=None,
        help=help_wrap(
            "The Cloud TPU to use for training. This should be either the name "
            "used when creating the Cloud TPU, or a "
            "grpc://ip.address.of.tpu:8470 url. Passing `local` will use the"
            "CPU of the local instance instead. (Good for debugging.)"))
    key_flags.append("tpu")

    flags.DEFINE_string(
        name="tpu_zone", default=None,
        help=help_wrap(
            "[Optional] GCE zone where the Cloud TPU is located in. If not "
            "specified, we will attempt to automatically detect the GCE "
            "project from metadata."))

    flags.DEFINE_string(
        name="tpu_gcp_project", default=None,
        help=help_wrap(
            "[Optional] Project name for the Cloud TPU-enabled project. If not "
            "specified, we will attempt to automatically detect the GCE "
            "project from metadata."))

    flags.DEFINE_integer(name="num_tpu_shards", default=8,
                         help=help_wrap("Number of shards (TPU chips)."))

  return key_flags



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Transformer-based text encoder network."""

from __future__ import absolute_import
from __future__ import division
# from __future__ import google_type_annotations
from __future__ import print_function

import inspect
import tensorflow as tf

from tensorflow.python.keras.engine import network  # pylint: disable=g-direct-tensorflow-import
from skydl.models_zoo.bert.official.nlp.modeling import layers


@tf.keras.utils.register_keras_serializable(package='Text')
class EncoderScaffold(network.Network):
  """Bi-directional Transformer-based encoder network scaffold.

  This network allows users to flexibly implement an encoder similar to the one
  described in "BERT: Pre-training of Deep Bidirectional Transformers for
  Language Understanding" (https://arxiv.org/abs/1810.04805).

  In this network, users can choose to provide a custom embedding subnetwork
  (which will replace the standard embedding logic) and/or a custom hidden layer
  class (which will replace the Transformer instantiation in the encoder). For
  each of these custom injection points, users can pass either a class or a
  class instance. If a class is passed, that class will be instantiated using
  the 'embedding_cfg' or 'hidden_cfg' argument, respectively; if an instance
  is passed, that instance will be invoked. (In the case of hidden_cls, the
  instance will be invoked 'num_hidden_instances' times.

  If the hidden_cls is not overridden, a default transformer layer will be
  instantiated.

  Attributes:
    num_output_classes: The output size of the classification layer.
    classification_layer_initializer: The initializer for the classification
      layer.
    classification_layer_dtype: The dtype for the classification layer.
    embedding_cls: The class or instance to use to embed the input data. This
      class or instance defines the inputs to this encoder. If embedding_cls is
      not set, a default embedding network (from the original BERT paper) will
      be created.
    embedding_cfg: A dict of kwargs to pass to the embedding_cls, if it needs to
      be instantiated. If embedding_cls is not set, a config dict must be
      passed to 'embedding_cfg' with the following values:
      "vocab_size": The size of the token vocabulary.
      "type_vocab_size": The size of the type vocabulary.
      "hidden_size": The hidden size for this encoder.
      "max_seq_length": The maximum sequence length for this encoder.
      "seq_length": The sequence length for this encoder.
      "initializer": The initializer for the embedding portion of this encoder.
      "dropout_rate": The dropout rate to apply before the encoding layers.
      "dtype": (Optional): The dtype of the embedding layers.
    embedding_data: A reference to the embedding weights that will be used to
      train the masked language model, if necessary. This is optional, and only
      needed if (1) you are overriding embedding_cls and (2) are doing standard
      pretraining.
    num_hidden_instances: The number of times to instantiate and/or invoke the
      hidden_cls.
    hidden_cls: The class or instance to encode the input data. If hidden_cls is
      not set, a KerasBERT transformer layer will be used as the encoder class.
    hidden_cfg: A dict of kwargs to pass to the hidden_cls, if it needs to be
      instantiated. If hidden_cls is not set, a config dict must be passed to
      'hidden_cfg' with the following values:
        "num_attention_heads": The number of attention heads. The hidden size
          must be divisible by num_attention_heads.
        "intermediate_size": The intermediate size of the transformer.
        "intermediate_activation": The activation to apply in the transfomer.
        "dropout_rate": The overall dropout rate for the transformer layers.
        "attention_dropout_rate": The dropout rate for the attention layers.
        "kernel_initializer": The initializer for the transformer layers.
        "dtype": The dtype of the transformer.
  """

  def __init__(
      self,
      num_output_classes,
      classification_layer_initializer=tf.keras.initializers.TruncatedNormal(
          stddev=0.02),
      classification_layer_dtype=tf.float32,
      embedding_cls=None,
      embedding_cfg=None,
      embedding_data=None,
      num_hidden_instances=1,
      hidden_cls=layers.Transformer,
      hidden_cfg=None,
      **kwargs):
    print(embedding_cfg)
    self._self_setattr_tracking = False
    self._hidden_cls = hidden_cls
    self._hidden_cfg = hidden_cfg
    self._num_hidden_instances = num_hidden_instances
    self._num_output_classes = num_output_classes
    self._classification_layer_initializer = classification_layer_initializer
    self._embedding_cls = embedding_cls
    self._embedding_cfg = embedding_cfg
    self._embedding_data = embedding_data
    self._kwargs = kwargs

    if embedding_cls:
      if inspect.isclass(embedding_cls):
        self._embedding_network = embedding_cls(embedding_cfg)
      else:
        self._embedding_network = embedding_cls
      inputs = self._embedding_network.inputs
      embeddings, mask = self._embedding_network(inputs)
    else:
      self._embedding_network = None
      word_ids = tf.keras.layers.Input(
          shape=(embedding_cfg['seq_length'],),
          dtype=tf.int32,
          name='input_word_ids')
      mask = tf.keras.layers.Input(
          shape=(embedding_cfg['seq_length'],),
          dtype=tf.int32,
          name='input_mask')
      type_ids = tf.keras.layers.Input(
          shape=(embedding_cfg['seq_length'],),
          dtype=tf.int32,
          name='input_type_ids')
      inputs = [word_ids, mask, type_ids]

      self._embedding_layer = layers.OnDeviceEmbedding(
          vocab_size=embedding_cfg['vocab_size'],
          embedding_width=embedding_cfg['hidden_size'],
          initializer=embedding_cfg['initializer'],
          name='word_embeddings')

      word_embeddings = self._embedding_layer(word_ids)

      # Always uses dynamic slicing for simplicity.
      self._position_embedding_layer = layers.PositionEmbedding(
          initializer=embedding_cfg['initializer'],
          use_dynamic_slicing=True,
          max_sequence_length=embedding_cfg['max_seq_length'])
      position_embeddings = self._position_embedding_layer(word_embeddings)

      type_embeddings = (
          layers.OnDeviceEmbedding(
              vocab_size=embedding_cfg['type_vocab_size'],
              embedding_width=embedding_cfg['hidden_size'],
              initializer=embedding_cfg['initializer'],
              use_one_hot=True,
              name='type_embeddings')(type_ids))

      embeddings = tf.keras.layers.Add()(
          [word_embeddings, position_embeddings, type_embeddings])
      embeddings = (
          tf.keras.layers.LayerNormalization(
              name='embeddings/layer_norm',
              axis=-1,
              epsilon=1e-12,
              dtype=tf.float32)(embeddings))
      embeddings = (
          tf.keras.layers.Dropout(
              rate=embedding_cfg['dropout_rate'], dtype=tf.float32)(embeddings))

      if embedding_cfg.get('dtype') == 'float16':
        embeddings = tf.cast(embeddings, tf.float16)

    attention_mask = layers.SelfAttentionMask()([embeddings, mask])
    data = embeddings

    for _ in range(num_hidden_instances):
      if inspect.isclass(hidden_cls):
        layer = self._hidden_cls(**hidden_cfg)
      else:
        layer = self._hidden_cls
      data = layer([data, attention_mask])

    first_token_tensor = (
        tf.keras.layers.Lambda(lambda x: tf.squeeze(x[:, 0:1, :], axis=1))(data)
    )
    cls_output = tf.keras.layers.Dense(
        units=num_output_classes,
        activation='tanh',
        kernel_initializer=classification_layer_initializer,
        dtype=classification_layer_dtype,
        name='cls_transform')(
            first_token_tensor)

    super(EncoderScaffold, self).__init__(
        inputs=inputs, outputs=[data, cls_output], **kwargs)

  def get_config(self):
    config_dict = {
        'num_hidden_instances':
            self._num_hidden_instances,
        'num_output_classes':
            self._num_output_classes,
        'classification_layer_initializer':
            self._classification_layer_initializer,
        'embedding_cls':
            self._embedding_network,
        'embedding_cfg':
            self._embedding_cfg,
        'hidden_cfg':
            self._hidden_cfg,
    }
    if inspect.isclass(self._hidden_cls):
      config_dict['hidden_cls_string'] = tf.keras.utils.get_registered_name(
          self._hidden_cls)
    else:
      config_dict['hidden_cls'] = self._hidden_cls

    config_dict.update(self._kwargs)
    return config_dict

  @classmethod
  def from_config(cls, config, custom_objects=None):
    if 'hidden_cls_string' in config:
      config['hidden_cls'] = tf.keras.utils.get_registered_object(
          config['hidden_cls_string'], custom_objects=custom_objects)
      del config['hidden_cls_string']
    return cls(**config)

  def get_embedding_table(self):
    if self._embedding_network is None:
      # In this case, we don't have a custom embedding network and can return
      # the standard embedding data.
      return self._embedding_layer.embeddings

    if self._embedding_data is None:
      raise RuntimeError(('The EncoderScaffold %s does not have a reference '
                          'to the embedding data. This is required when you '
                          'pass a custom embedding network to the scaffold. '
                          'It is also possible that you are trying to get '
                          'embedding data from an embedding scaffold with a '
                          'custom embedding network where the scaffold has '
                          'been serialized and deserialized. Unfortunately, '
                          'accessing custom embedding references after '
                          'serialization is not yet supported.') % self.name)
    else:
      return self._embedding_data



import os
import codecs
import skydl
from setuptools import setup, find_packages

with open("README.md") as f:
    readme = f.read()


def read(fname):
    return codecs.open(os.path.join(os.path.dirname(__file__), fname)).read()


def read_install_requires():
    with open('requirements.txt', 'r') as f:
        res = f.readlines()
    res = list(map(lambda s: s.replace('\n', ''), res))
    return res


setup(
    name='skydl',
    version=skydl.__version__,
    description="",
    long_description=readme,
    install_requires=read_install_requires(),
    setup_requires=['setuptools>=41.0.1', 'wheel>=0.33.4'],
    author='tony',
    author_email='',
    license='BSD',
    url='',
    keywords=['robot', 'ai', 'reinforcement learning', 'machine learning', 'RL',
              'ML', 'tensorflow', 'pytorch', 'ray', 'skydl', 'high efficiency', 'deep learning'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Framework :: Robot Framework :: Library',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'],
    packages=find_packages(exclude=[]),
    package_data={'': []},
)


# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Trainer network for BERT-style models."""

from __future__ import absolute_import
from __future__ import division
# from __future__ import google_type_annotations
from __future__ import print_function

import tensorflow as tf

from skydl.models_zoo.bert.official.nlp.modeling import networks


@tf.keras.utils.register_keras_serializable(package='Text')
class BertSpanLabeler(tf.keras.Model):
  """Span labeler model based on a BERT-style transformer-based encoder.

  This is an implementation of the network structure surrounding a transformer
  encoder as described in "BERT: Pre-training of Deep Bidirectional Transformers
  for Language Understanding" (https://arxiv.org/abs/1810.04805).

  The BertSpanLabeler allows a user to pass in a transformer stack, and
  instantiates a span labeling network based on a single dense layer.

  Attributes:
    network: A transformer network. This network should output a sequence output
      and a classification output. Furthermore, it should expose its embedding
      table via a "get_embedding_table" method.
    initializer: The initializer (if any) to use in the span labeling network.
      Defaults to a Glorot uniform initializer.
    output: The output style for this network. Can be either 'logits' or
      'predictions'.
  """

  def __init__(self,
               network,
               initializer='glorot_uniform',
               output='logits',
               **kwargs):
    self._self_setattr_tracking = False
    self._config = {
        'network': network,
        'initializer': initializer,
        'output': output,
    }
    # We want to use the inputs of the passed network as the inputs to this
    # Model. To do this, we need to keep a handle to the network inputs for use
    # when we construct the Model object at the end of init.
    inputs = network.inputs

    # Because we have a copy of inputs to create this Model object, we can
    # invoke the Network object with its own input tensors to start the Model.
    sequence_output, _ = network(inputs)

    # This is an instance variable for ease of access to the underlying task
    # network.
    self.span_labeling = networks.SpanLabeling(
        input_width=sequence_output.shape[-1],
        initializer=initializer,
        output=output,
        name='span_labeling')
    start_logits, end_logits = self.span_labeling(sequence_output)

    # Use identity layers wrapped in lambdas to explicitly name the output
    # tensors. This allows us to use string-keyed dicts in Keras fit/predict/
    # evaluate calls.
    start_logits = tf.keras.layers.Lambda(
        tf.identity, name='start_positions')(
            start_logits)
    end_logits = tf.keras.layers.Lambda(
        tf.identity, name='end_positions')(
            end_logits)

    logits = [start_logits, end_logits]

    super(BertSpanLabeler, self).__init__(
        inputs=inputs, outputs=logits, **kwargs)

  def get_config(self):
    return self._config

  @classmethod
  def from_config(cls, config, custom_objects=None):
    return cls(**config)



# -*- coding: utf-8 -*-
import tensorflow as tf
from tensorflow import keras
from skydl.models.netv2 import NetV2
from skydl.models.tf_keras_modelv2_mixin import TfKerasModelV2Mixin


class TfKerasNetV2(NetV2, TfKerasModelV2Mixin):
    """
    default tensorflow2 keras net implements
    """
    @property
    def model_proxy(self)->keras.Model:
        """model_proxy maybe None or keras.Model"""
        return self._model_proxy

    def __init__(self, name=None, parser_args=None, keras_layers=[], keras_inputs=None, keras_outputs=None):
        # super()会默认调用第一个父类(SuperNet)的__init__()
        super().__init__(name=name, parser_args=parser_args)
        # 再显式调用其它的__init__(self)
        TfKerasModelV2Mixin.__init__(self, keras_layers, keras_inputs=keras_inputs, keras_outputs=keras_outputs)
        # 构建keras.Model代理类接受keras_inputs, keras_outputs参数
        if keras_inputs is not None and keras_outputs is not None:
            self._model_proxy = keras.Model(inputs=keras_inputs, outputs=keras_outputs)

    def forward(self, *input, **kwargs):
        return super().forward(*input, **kwargs)

    def call(self, _x: tf.Tensor=None, training: bool=True) -> tf.Tensor:
        return self.forward(_x)

    def call_hidden_layers(self, _x: tf.Tensor=None, name="prediction"):
        if self._model_proxy is None:  # self is subclass of keras.Sequential
            return self.call_layers(_x, name)
        else:
            return self._model_proxy.call(_x)

    def compile_loss_metrics_optimizer_predict(self, _x, _y, _learning_rate=0.001, devices=[]):
        pass





# -*- coding: utf-8 -*-
import ray
import logging
import traceback
import time
from py_common_util.common.date_utils import DateUtils
from skydl.ray.streaming_0_8_2.python.streaming import Environment, DataStream
from skydl.ray.ray_streaming_util import DefaultStreamingSource, DefaultStreamingEnvConfig, \
    DefaultStreamingEnvConfigV082

# define logger
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class RayStreamingExecutor:
    """
    ray流计算执行器，子类可以重写build_stream()方法
    """
    @property
    def args_parser(self):
        return self._args_parser

    def __init__(self, args_parser=None):
        self._args_parser = args_parser
        logger.info("starting a streaming executor...")

    def execute(self, env=None):
        try:
            if not env:
                env = Environment(DefaultStreamingEnvConfigV082(parallelism=self.args_parser.streaming_parallelism))
            stream = self.build_stream(env=env)
            self.execute_forever(env, stream)
        except Exception as e:
            logger.error("RayStreamingExecutor fail on execute(): %s", str(e))
            traceback.print_exc()

    def execute_forever(self, env, stream, loop_idle_seconds=1):
        while(True):
            try:
                # env.print_logical_graph()
                ray.get(env.execute())
                logger.info("Ray Output stream id: {}".format(stream.id))
            except Exception as e:
                logger.error("Error Occurred on Ray Streaming: %s", str(e))
                traceback.print_exc()
                # TODO 发送异常运维警报
            finally:
                time.sleep(loop_idle_seconds)  # 等待所有算子的数据都执行完，然后休息1秒后再继续重新执行同一个stream计算图

    def build_stream(self, env)->DataStream:
        args_parser = self.args_parser
        """
        子类需要重写该方法构建具体的业务stream
        注意streaming算子类必须要预先install在所有节点中，或者都在该方法内部定义，尝试了importlib的方式会报错
        @:param args_parser
        @:param env
        @:return DataStream
        """
        class FooStreamingSource(DefaultStreamingSource):
            def __init__(self):
                logger.info("FooStreamingSource init...(init func only be executed 1 time in a stream graph!!!)")
                self._count = 0

            def get_next(self):
                self._count += 1
                logger.info("get_next:" + str(self._count))
                if self._count % 2 == 0:
                    time.sleep(5)  # with env config background_flush=True, even if it sleep ever, it also success to auto flush。but can not kill current stream until return None。
                    # return None
                    return str(self._count) + "-aaa\nbbb*"+DateUtils.now_to_str()
                else:
                    return str(self._count) + "-ccc*"+DateUtils.now_to_str()

        def splitter_fn(record):
            """convert record to another record list, e.g. [record]"""
            logger.info("splitter: " + record)
            # return [record+"-splitter-" + str(i) for i in range(20)]
            return [record+"-splitter-" + DateUtils.now_to_str()]

        def handle_record_fn(record):
            """处理回测的逻辑"""
            logger.info("args_parser=" + str(args_parser))
            logger.info("record=" + record)
            time.sleep(10)
            logger.info(record + ", finished:" + DateUtils.now_to_str())

        stream = env.source(
            FooStreamingSource()
        ).round_robin().set_parallelism(2).flat_map(
            flatmap_fn=splitter_fn
        ).set_parallelism(1).map(
            map_fn=handle_record_fn,
            name="nocode_backtest_map"
        ).set_parallelism(2)
        return stream













# -*- coding: utf-8 -*-
import tensorflow as tf
from skydl.models.netv2 import NetV2
from skydl.models.estimator_modelv2_mixin import EstimatorModelV2Mixin


class EstimatorNetV2(NetV2, EstimatorModelV2Mixin):
    """
    default tf estimator net wrapper
    """
    def __init__(self, name=None, parser_args=None, layers=[], model_fn=None, params=None):
        # super()会默认调用第一个父类(SuperNet)的__init__()
        super().__init__(name=name, parser_args=parser_args)
        # 再显式调用其它的__init__(self)
        EstimatorModelV2Mixin.__init__(self, layers, model_fn, params=params)

    def forward(self, *input, **kwargs):
        return super().forward(*input, **kwargs)

    def call(self, _x: tf.Tensor=None, training: bool=True) -> tf.Tensor:
        return self.forward(_x)

    def call_hidden_layers(self, _x: tf.Tensor=None, name="prediction"):
        return self.call_layers(_x, name)

    def compile_loss_metrics_optimizer_predict(self, _x, _y, _learning_rate=0.001, devices=[]):
        pass





# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Helper functions for the Keras implementations of models."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import multiprocessing
import os
import time

from absl import logging
import tensorflow as tf
from tensorflow.core.protobuf import rewriter_config_pb2
from tensorflow.python import tf2
from tensorflow.python.eager import profiler


class BatchTimestamp(object):
  """A structure to store batch time stamp."""

  def __init__(self, batch_index, timestamp):
    self.batch_index = batch_index
    self.timestamp = timestamp

  def __repr__(self):
    return "'BatchTimestamp<batch_index: {}, timestamp: {}>'".format(
        self.batch_index, self.timestamp)


class TimeHistory(tf.keras.callbacks.Callback):
  """Callback for Keras models."""

  def __init__(self, batch_size, log_steps):
    """Callback for logging performance.

    Args:
      batch_size: Total batch size.
      log_steps: Interval of steps between logging of batch level stats.
    """
    self.batch_size = batch_size
    super(TimeHistory, self).__init__()
    self.log_steps = log_steps
    self.global_steps = 0

    # Logs start of step 1 then end of each step based on log_steps interval.
    self.timestamp_log = []

    # Records the time each epoch takes to run from start to finish of epoch.
    self.epoch_runtime_log = []

  def on_train_end(self, logs=None):
    self.train_finish_time = time.time()

  def on_epoch_begin(self, epoch, logs=None):
    self.epoch_start = time.time()

  def on_batch_begin(self, batch, logs=None):
    self.global_steps += 1
    if self.global_steps == 1:
      self.start_time = time.time()
      self.timestamp_log.append(BatchTimestamp(self.global_steps,
                                               self.start_time))

  def on_batch_end(self, batch, logs=None):
    """Records elapse time of the batch and calculates examples per second."""
    if self.global_steps % self.log_steps == 0:
      timestamp = time.time()
      elapsed_time = timestamp - self.start_time
      examples_per_second = (self.batch_size * self.log_steps) / elapsed_time
      self.timestamp_log.append(BatchTimestamp(self.global_steps, timestamp))
      logging.info(
          "BenchmarkMetric: {'global step':%d, 'time_taken': %f,"
          "'examples_per_second': %f}",
          self.global_steps, elapsed_time, examples_per_second)
      self.start_time = timestamp

  def on_epoch_end(self, epoch, logs=None):
    epoch_run_time = time.time() - self.epoch_start
    self.epoch_runtime_log.append(epoch_run_time)
    logging.info(
        "BenchmarkMetric: {'epoch':%d, 'time_taken': %f}",
        epoch, epoch_run_time)


def get_profiler_callback(model_dir, profile_steps, enable_tensorboard,
                          steps_per_epoch):
  """Validate profile_steps flag value and return profiler callback."""
  profile_steps_error_message = (
      'profile_steps must be a comma separated pair of positive integers, '
      'specifying the first and last steps to be profiled.'
  )
  try:
    profile_steps = [int(i) for i in profile_steps.split(',')]
  except ValueError:
    raise ValueError(profile_steps_error_message)
  if len(profile_steps) != 2:
    raise ValueError(profile_steps_error_message)
  start_step, stop_step = profile_steps
  if start_step < 0 or start_step > stop_step:
    raise ValueError(profile_steps_error_message)
  if enable_tensorboard:
    logging.warning(
        'Both TensorBoard and profiler callbacks are used. Note that the '
        'TensorBoard callback profiles the 2nd step (unless otherwise '
        'specified). Please make sure the steps profiled by the two callbacks '
        'do not overlap.')
  return ProfilerCallback(model_dir, start_step, stop_step, steps_per_epoch)


class ProfilerCallback(tf.keras.callbacks.Callback):
  """Save profiles in specified step range to log directory."""

  def __init__(self, log_dir, start_step, stop_step, steps_per_epoch):
    super(ProfilerCallback, self).__init__()
    self.log_dir = log_dir
    self.start_step = start_step
    self.stop_step = stop_step
    self.start_epoch = start_step // steps_per_epoch
    self.stop_epoch = stop_step // steps_per_epoch
    self.start_step_in_epoch = start_step % steps_per_epoch
    self.stop_step_in_epoch = stop_step % steps_per_epoch
    self.should_start = False
    self.should_stop = False

  def on_epoch_begin(self, epoch, logs=None):
    if epoch == self.start_epoch:
      self.should_start = True
    if epoch == self.stop_epoch:
      self.should_stop = True

  def on_batch_begin(self, batch, logs=None):
    if batch == self.start_step_in_epoch and self.should_start:
      self.should_start = False
      profiler.start()
      logging.info('Profiler started at Step %s', self.start_step)

  def on_batch_end(self, batch, logs=None):
    if batch == self.stop_step_in_epoch and self.should_stop:
      self.should_stop = False
      results = profiler.stop()
      profiler.save(self.log_dir, results)
      logging.info(
          'Profiler saved profiles for steps between %s and %s to %s',
          self.start_step, self.stop_step, self.log_dir)


def set_session_config(enable_eager=False,
                       enable_xla=False):
  """Sets the session config."""
  if is_v2_0():
    set_config_v2(enable_xla=enable_xla)
  else:
    config = get_config_proto_v1(enable_xla=enable_xla)
    if enable_eager:
      tf.compat.v1.enable_eager_execution(config=config)
    else:
      sess = tf.Session(config=config)
      tf.keras.backend.set_session(sess)


def get_config_proto_v1(enable_xla=False):
  """Return config proto according to flag settings, or None to use default."""
  config = None
  if enable_xla:
    config = tf.compat.v1.ConfigProto()
    config.graph_options.optimizer_options.global_jit_level = (
        tf.OptimizerOptions.ON_2)
  return config


def set_config_v2(enable_xla=False):
  """Config eager context according to flag values using TF 2.0 API."""
  if enable_xla:
    tf.config.optimizer.set_jit(True)


def is_v2_0():
  """Returns true if using tf 2.0."""
  return tf2.enabled()


def set_gpu_thread_mode_and_count(gpu_thread_mode,
                                  datasets_num_private_threads,
                                  num_gpus, per_gpu_thread_count):
  """Set GPU thread mode and count, and adjust dataset threads count."""
  cpu_count = multiprocessing.cpu_count()
  logging.info('Logical CPU cores: %s', cpu_count)

  # Allocate private thread pool for each GPU to schedule and launch kernels
  per_gpu_thread_count = per_gpu_thread_count or 2
  os.environ['TF_GPU_THREAD_MODE'] = gpu_thread_mode
  os.environ['TF_GPU_THREAD_COUNT'] = str(per_gpu_thread_count)
  logging.info('TF_GPU_THREAD_COUNT: %s',
               os.environ['TF_GPU_THREAD_COUNT'])
  logging.info('TF_GPU_THREAD_MODE: %s',
               os.environ['TF_GPU_THREAD_MODE'])

  # Limit data preprocessing threadpool to CPU cores minus number of total GPU
  # private threads and memory copy threads.
  total_gpu_thread_count = per_gpu_thread_count * num_gpus
  num_runtime_threads = num_gpus
  if not datasets_num_private_threads:
    datasets_num_private_threads = min(
        cpu_count - total_gpu_thread_count - num_runtime_threads,
        num_gpus * 8)
    logging.info('Set datasets_num_private_threads to %s',
                 datasets_num_private_threads)
